/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.model.memory.loader;

import ru.ispras.castle.util.Logger;

import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.data.types.bitvector.BitVectorMath;
import ru.ispras.microtesk.model.Model;
import ru.ispras.microtesk.model.ProcessingElement;
import ru.ispras.microtesk.model.memory.MemoryDevice;
import ru.ispras.microtesk.utils.BitStream;

import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Path;
import java.nio.file.Files;
import java.util.List;
import java.util.Set;

/**
 * @author <a href="mailto:kotsynyak@ispras.ru">Artem Kotsynyak</a>
 */
public class MemoryImageLoader {
  public static final class Request {
    private final String targetId;
    private final long targetOffset;
    private final long targetLength;

    private final InputStream source;

    private Request(String targetId, long offset, long length, InputStream source) {
      this.targetId = targetId;
      this.targetOffset = offset;
      this.targetLength = length;
      this.source = source;
    }
  }

  public static Request request(
      String target, long targetLo, long targetHi, Path path, long sourceLo) throws IOException {
    final var source = newPositionedInputStream(path, sourceLo);
    return new Request(target, targetLo, targetHi - targetLo, source);
  }

  public static Request request(
      String target, long targetLo, Path path, long sourceLo, long sourceHi) throws IOException {
    final var source = newPositionedInputStream(path, sourceLo);
    return new Request(
        target, targetLo, (sourceHi > 0) ? sourceHi - sourceLo : -1, source);
  }

  public static void processRequests(
      Model model, List<Request> reqs) throws IOException {
    final ProcessingElement pe = model.getPE();
    final Set<String> knownMem = pe.getMemoryStorageSet();
    for (Request req: reqs) {
      if (knownMem.contains(req.targetId)) {
        final var mem = pe.getMemoryDevice(req.targetId);
        final var task = new CopyTask(mem, req);

        if (req.targetLength >= 0) {
          task.copyRange(req.targetLength);
        } else {
          task.copyAll();
        }
      } else {
        Logger.warning(
            "Request to load an image into an inexistent memory '%s', ignoring the request",
            req.targetId);
      }

      req.source.close();
    }
  }

  private static final class CopyTask {
    private final BitStream source;
    private final int readSize;

    private final MemoryDevice target;
    private final BitVector addrOffset;
    private BitVector address;

    private CopyTask(MemoryDevice target, Request req) {
      this.source = BitStream.of(req.source);
      this.readSize = target.getDataBitSize();

      final int addrSize = target.getAddressBitSize();
      this.target = target;
      this.addrOffset = BitVector.valueOf(1, addrSize);
      this.address = BitVector.valueOf(req.targetOffset, addrSize);
    }

    void copyAll() throws IOException {
      while (!source.isEmpty()) {
        copyNext();
      }
    }

    void copyRange(long n) throws IOException {
      for (long i = 0; i < n; ++i) {
        copyNext();
      }
    }

    void copyNext() throws IOException {
      BitVector value = source.read(readSize);
      target.store(address, 0, value);
      address = BitVectorMath.add(address, addrOffset);
    }
  }

  private static InputStream newPositionedInputStream(
      Path path, long offset) throws IOException {
    final var source = Files.newInputStream(path);

    long nskipped = 0;
    while (nskipped < offset) {
      long rc = source.skip(offset - nskipped);
      if (rc == 0) {
        throw new java.io.EOFException();
      }
      nskipped += rc;
    }
    return source;
  }
}
