/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.model.memory.loader;

import javax.json.JsonArray;
import javax.json.JsonObject;
import javax.json.JsonReader;
import javax.json.JsonValue;

import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.model.Model;
import ru.ispras.microtesk.model.ProcessingElement;
import ru.ispras.microtesk.model.memory.MemoryDevice;
import ru.ispras.microtesk.utils.JsonConfig;

/**
 * The {@link MemoryLoader} class is implements loading of data from the file into the model's
 * memory.
 * 
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class MemoryLoader {

  private JsonReader jsonReader;

  private MemoryLoader(final String fileName) {
    InvariantChecks.checkNotNull(fileName);

    jsonReader = JsonConfig.getJsonReader(fileName);
  }

  private void close() {
    InvariantChecks.checkNotNull(jsonReader);

    JsonConfig.close(jsonReader);
  }

  private void loadMemoryElements(final Model model) {
    InvariantChecks.checkNotNull(model);
    ProcessingElement pe = model.getPE();

    JsonArray jsonArray = jsonReader.readArray();
    for (JsonValue jsonValue : jsonArray) {
      JsonObject jsonObject = jsonValue.asJsonObject();

      final String memoryName = jsonObject.getString(MemoryRegion.TAG_MEMORY_NAME);

      final MemoryDevice memoryDevice = pe.getMemoryDevice(memoryName);
      final int addressBitSize = memoryDevice.getAddressBitSize();
      final int dataSize = memoryDevice.getDataBitSize();

      final int address =
          Integer.parseInt(jsonObject.getString(MemorySaver.TAG_MEMORY_ADDRESS), 16);
      final int value = Integer.parseInt(jsonObject.getString(MemorySaver.TAG_MEMORY_VALUE), 16);

      // test: address + 12
      memoryDevice.store(BitVector.valueOf(address, addressBitSize),
          BitVector.valueOf(value, dataSize));

      // BitVector temp = memoryDevice.load(BitVector.valueOf(address+512, addressBitSize));
      // System.out.println(String.format("upd: address: %s, value: %s", address+512,
      // temp.intValue()));
    }
  }

  public static void loadMemoryElement(final MemoryDevice memoryDevice, final int address,
      final int value) {
    final int addressBitSize = memoryDevice.getAddressBitSize();
    final int dataSize = memoryDevice.getDataBitSize();

    memoryDevice.store(BitVector.valueOf(address, addressBitSize),
        BitVector.valueOf(value, dataSize));
  }

  public static void loadMemoryElement(final MemoryDevice memoryDevice, final int address,
      final BitVector value) {
    final int addressBitSize = memoryDevice.getAddressBitSize();

    memoryDevice.store(BitVector.valueOf(address, addressBitSize), value);
  }

  public static void loadMemoryElements(final String fileName, final Model model) {
    final MemoryLoader memoryWriterElements = new MemoryLoader(fileName);
    memoryWriterElements.loadMemoryElements(model);
    memoryWriterElements.close();
  }
}
