/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.model.memory.loader;

import java.util.HashSet;
import java.util.Set;

import javax.json.JsonArray;
import javax.json.JsonObject;
import javax.json.JsonReader;
import javax.json.JsonValue;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.utils.JsonConfig;

/**
 * The {@link MemoryRegion} class contains information about memory region for loading and writing
 * data.
 * 
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class MemoryRegion {
  public static final String TAG_MEMORY_NAME    = "memory_name";
  public static final String TAG_START_INDEX    = "start_index";
  public static final String TAG_END_INDEX      = "end_index";
  public static final String TAG_STRUCTURE_TYPE = "structure_type";

  final private String memoryName;
  final private int startIndex;
  final private int endIndex;
  final private String type;

  MemoryRegion(final String memoryName, final int start, final int end, final String type) {
    this.memoryName = memoryName;
    this.startIndex = start;
    this.endIndex = end;
    this.type = type;
  }

  public String getMemoryName() {
    return memoryName;
  }

  public int getStartIndex() {
    return startIndex;
  }

  public int getEndIndex() {
    return endIndex;
  }

  public String getType() {
    return type;
  }

  @Override
  public String toString() {
    return String.format("memory name: '%s', start index: 0x%s, end index: 0x%s, type: '%s'",
        memoryName, Integer.toHexString(startIndex), Integer.toHexString(endIndex), type);
  }

  public static Set<MemoryRegion> loadMemoryRegions(final String fileName) {
    InvariantChecks.checkNotNull(fileName);
    JsonReader jsonReader = JsonConfig.getJsonReader(fileName);

    Set<MemoryRegion> regions = new HashSet<>();

    JsonArray jsonArray = jsonReader.readArray();
    for (JsonValue jsonValue : jsonArray) {
      JsonObject jsonObject = jsonValue.asJsonObject();

      final String memoryName = jsonObject.getString(MemoryRegion.TAG_MEMORY_NAME);
      final int startIndax = Integer.parseInt(jsonObject.getString(MemoryRegion.TAG_START_INDEX), 16);
      final int endIndex = Integer.parseInt(jsonObject.getString(MemoryRegion.TAG_END_INDEX), 16);
      final String structureType = jsonObject.getString(MemoryRegion.TAG_STRUCTURE_TYPE);

      MemoryRegion memoryRegion = new MemoryRegion(memoryName, startIndax, endIndex, structureType);
      Logger.debug("Memory region loaded: [%s]", memoryRegion.toString());

      regions.add(memoryRegion);
    }

    JsonConfig.close(jsonReader);

    return regions;
  }
}
