/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.model.memory.loader;

import java.math.BigInteger;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;
import java.util.Set;

import javax.json.Json;
import javax.json.JsonArrayBuilder;
import javax.json.JsonObjectBuilder;
import javax.json.JsonWriter;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.model.Model;
import ru.ispras.microtesk.model.ProcessingElement;
import ru.ispras.microtesk.model.memory.MemoryDevice;
import ru.ispras.microtesk.utils.JsonConfig;

/**
 * The {@link MemorySaver} class is implements saving model memory to a file
 * 
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class MemorySaver {
  public static final String TAG_MEMORY_ADDRESS = "address";
  public static final String TAG_MEMORY_VALUE = "value";
  public static final String TAG_MEMORY_ELEMENT = "element";

  public static final String TAG_DATA_SIZE = "data_size";
  public static final String TAG_ADDRESS_BIT_SIZE = "address_bit_size";

  // TODO: Empty memory check
  public static final int MEMORY_EMPTY_DEPTH = 2;

  private JsonWriter jsonWriter;

  private MemorySaver(final String fileName) {
    InvariantChecks.checkNotNull(fileName);

    jsonWriter = JsonConfig.getJsonWriter(fileName);
  }

  private boolean saveMemoryStoragesList(final ProcessingElement pe) {
    InvariantChecks.checkNotNull(pe);

    Set<String> memoryStorageNameSet = pe.getMemoryStorageSet();

    final JsonArrayBuilder jsonArrayBuilder = Json.createArrayBuilder();
    for (final String memoryStorageName : memoryStorageNameSet) {
      final JsonObjectBuilder objectBuilder = Json.createObjectBuilder();
      objectBuilder.add(MemoryRegion.TAG_MEMORY_NAME, memoryStorageName);
      {
        final MemoryDevice memoryDevice = pe.getMemoryDevice(memoryStorageName);
        final int addressBitSize = memoryDevice.getAddressBitSize();
        final int dataSize = memoryDevice.getDataBitSize();
        objectBuilder.add(MemorySaver.TAG_ADDRESS_BIT_SIZE, addressBitSize);
        objectBuilder.add(MemorySaver.TAG_DATA_SIZE, dataSize);
      }

      Logger.debug("Storage '%s' has been added to the file.", memoryStorageName);
      jsonArrayBuilder.add(objectBuilder);
    }
    jsonWriter.writeArray(jsonArrayBuilder.build());
    return true;
  }

  private void close() {
    InvariantChecks.checkNotNull(jsonWriter);

    JsonConfig.close(jsonWriter);
  }

  public static void saveProcessingElementStorage(final String fileName, final Model model) {
    ProcessingElement pe = model.getPE();
    InvariantChecks.checkNotNull(pe);

    Logger.debug("Save to file '%s' the ProcessingElement storage(%s) list.", fileName,
        pe.getMemoryStorageSet().size());

    MemorySaver memoryWriter = new MemorySaver(fileName);
    memoryWriter.saveMemoryStoragesList(pe);
    memoryWriter.close();
  }

  public static void saveMemoryElements(final Model model) {
    InvariantChecks.checkNotNull(model);
    ProcessingElement pe = model.getPE();

    final Path fileRegions = ParametersLoader.getRegionsPath(model.getName());
    final Path fileStructure = ParametersLoader.getStructurePath(model.getName());
    if (Files.exists(fileRegions) && Files.exists(fileStructure)) {
      final Set<MemoryRegion> memoryRegions =
          MemoryRegion.loadMemoryRegions(fileRegions.toString());

      final Map<String, MemoryStructure> memoryStructure =
          MemoryStructure.loadMemoryElementStructure(fileStructure.toString());

      final MemorySaver memoryWriterElements = new MemorySaver("MemoryElements.json");
      memoryWriterElements.saveMemoryElements(pe, memoryRegions, memoryStructure);
      memoryWriterElements.close();
    } else {
      Logger.debug("The file '%s' exist: %s", fileRegions, Files.exists(fileRegions));
      Logger.debug("The file '%s' exist: %s", fileStructure, Files.exists(fileStructure));
      return;
    }
  }

  private boolean saveMemoryElements(final ProcessingElement pe,
      final Set<MemoryRegion> memoryRegions, final Map<String, MemoryStructure> memoryStructure) {
    InvariantChecks.checkNotNull(pe);
    InvariantChecks.checkNotNull(memoryRegions);
    InvariantChecks.checkNotNull(memoryStructure);

    final JsonArrayBuilder jsonArrayBuilder = Json.createArrayBuilder();
    for (final MemoryRegion memoryRegion : memoryRegions) {
      Logger.debug("Reading a memory region: [%s]", memoryRegion);

      final String memoryName = memoryRegion.getMemoryName();
      final int address = memoryRegion.getStartIndex();
      final int cycleEnd = memoryRegion.getEndIndex() - address;
      final String structureType = memoryRegion.getType();

      final MemoryDevice memoryDevice = pe.getMemoryDevice(memoryName);
      final int dataSize = memoryDevice.getDataBitSize();

      MemoryStructure structure = memoryStructure.get(structureType);
      final String[] elements = structure.getElements();

      BigInteger emptyCheck = BigInteger.ZERO;
      boolean cycleStop = false;
      int cycleEmptyDepth = 0;

      for (int i = 0; i < cycleEnd && !cycleStop; i++) {
        final JsonObjectBuilder objectBuilder = Json.createObjectBuilder();
        final int index = address + i;

        objectBuilder.add(MemoryRegion.TAG_MEMORY_NAME, memoryName);
        objectBuilder.add(MemorySaver.TAG_MEMORY_ADDRESS, Integer.toHexString(index));

        final BitVector value = memoryDevice.load(BitVector.valueOf(index, dataSize));
        emptyCheck = emptyCheck.add(value.bigIntegerValue(false));
        objectBuilder.add(MemorySaver.TAG_MEMORY_VALUE, value.toHexString());

        objectBuilder.add(MemoryRegion.TAG_STRUCTURE_TYPE, structureType);
        objectBuilder.add(MemorySaver.TAG_MEMORY_ELEMENT, elements[i % elements.length]);

        Logger.debug("Adds to Json: %s, type: %s, address: %s, value: %s, element: %s", memoryName,
            structureType, Integer.toHexString(index), value.toHexString(),
            elements[i % elements.length]);

        // Condition to not save empty memory
        if (i % elements.length == elements.length - 1) {
          if (!emptyCheck.equals(BigInteger.ZERO)) {
            Logger.debug("Save to file the memory json object: %s, type: %s, address: %s",
                memoryName, structureType, Integer.toHexString(index));

            emptyCheck = BigInteger.ZERO;
            cycleEmptyDepth = 0;
          } else {
            cycleEmptyDepth++;
            Logger.debug("Empty memory block: %s", emptyCheck);
          }
          if (cycleEmptyDepth == MemorySaver.MEMORY_EMPTY_DEPTH) {
            cycleStop = true;
          }
        }

        jsonArrayBuilder.add(objectBuilder);
      }
    }
    jsonWriter.writeArray(jsonArrayBuilder.build());

    return true;
  }
}
