/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.model.memory.loader;

import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import javax.json.JsonArray;
import javax.json.JsonObject;
import javax.json.JsonReader;
import javax.json.JsonString;
import javax.json.JsonValue;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.utils.JsonConfig;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class MemoryStructure {
  public static final String TYPE = "structure_type";
  public static final String SIZE = "structure_size";
  public static final String ELEMENTS = "elements";
  public static final String DATA_BIT_SIZE = "data_bit_size";

  final private String type;
  final private int elementSize;
  final private int dataSize;
  final private String[] elements;

  MemoryStructure(final String type, final int structureSize, final String[] elements,
      final int dataSize) {
    this.type = type;
    this.elementSize = structureSize;
    this.elements = elements;
    this.dataSize = dataSize;
  }

  public String getType() {
    return type;
  }

  public int getSize() {
    return elementSize;
  }

  public String[] getElements() {
    return elements;
  }

  public int getDataSize() {
    return dataSize;
  }

  @Override
  public String toString() {
    return String.format("type: '%s', size: %s, elements: %s, data bit size: %s", type, elementSize,
        Arrays.toString(elements), dataSize);
  }

  public static Map<String, MemoryStructure> loadMemoryElementStructure(final String fileName) {
    InvariantChecks.checkNotNull(fileName);
    JsonReader jsonReader = JsonConfig.getJsonReader(fileName);

    Map<String, MemoryStructure> regions = new HashMap<>();

    JsonArray jsonArray = jsonReader.readArray();
    for (JsonValue jsonValue : jsonArray) {
      JsonObject jsonObject = jsonValue.asJsonObject();

      final String type = jsonObject.getString(MemoryStructure.TYPE);
      final int elementSize = jsonObject.getInt(MemoryStructure.SIZE);
      final List<JsonString> elementsJsonList =
          jsonObject.getJsonArray(MemoryStructure.ELEMENTS).getValuesAs(JsonString.class);
      final int dataSize = jsonObject.getInt(MemoryStructure.DATA_BIT_SIZE);

      final List<String> elementsList =
          elementsJsonList.stream().map(JsonString::getString).collect(Collectors.toList());

      final String[] elements = elementsList.toArray(arraySize -> new String[arraySize]);

      MemoryStructure memoryStructure = new MemoryStructure(type, elementSize, elements, dataSize);
      Logger.debug("Memory structure loaded: [%s]", memoryStructure.toString());

      regions.put(memoryStructure.getType(), memoryStructure);
    }

    JsonConfig.close(jsonReader);
    return regions;
  }
}
