/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.model.tracer;

import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.test.template.ConcreteCall;

/**
 * The {@link ArkRecord} class describes Ark(Panda) Tracer log records.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public abstract class ArkRecord extends Record {
  private static long instructionId = -1;

  // Thread identification
  protected int tid;

  public ArkRecord(final RecordKind kind, final long time) {
    super(kind, time);
  }

  public static void resetInstructionCount() {
    instructionId = -1;
  }

  @Override
  public String toString() {
    return String.format("%d ", this.getTime());
  }

  /**
   * Returns an object {@link ArkRecord} for the log tracer the Ark(Panda) instruction.
   *
   * @param call Program instructions.
   * @param tid Thread identification in which the instruction was executed.
   *
   * @return New instance of the {@link ArkRecord} for instruction.
   */
  public static Record newInstruction(final ConcreteCall call, final int tid) {
    return new Instruction(call, tid);
  }

  public static Record newRegisterWrite(final String register, final BitVector value) {
    return new RegisterWrite(register, value);
  }

  private static final class Instruction extends ArkRecord {
    private long addr;
    private String disasm;

    private Instruction(final ConcreteCall call, final int tid) {
      super(RecordKind.INSTRUCT, ++instructionId);
      InvariantChecks.checkNotNull(call);

      this.tid = tid;
      this.addr = call.getAddress().longValue();
      this.disasm = call.getExecutable().getText();
    }

    @Override
    public String toString() {
      return String.format("%s [TID %d] D/interpreter: %s: %s",
          super.toString(),
          tid,
          Long.toHexString(addr),
          disasm);
    }
  }

  private static final class RegisterWrite extends ArkRecord {
    private final String register;
    private final BitVector value;

    private RegisterWrite(final String register, final BitVector value) {
      super(RecordKind.REGISTER, instructionId);

      InvariantChecks.checkNotNull(register);
      InvariantChecks.checkNotNull(value);

      this.register = register;
      this.value = value;
    }

    @Override
    public String toString() {
      return String.format("%s [TID %d] D/interpreter: %s%s = (hex) %s",
          super.toString(),
          tid,
          "        ", // shift by 8
          register,
          value.toHexString().toLowerCase());
    }
  }
}
