/*
 * Copyright 2022-2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.adapter;

import java.io.File;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.List;

import ru.ispras.castle.util.FileUtils;
import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.options.Options;
import ru.ispras.microtesk.tools.ToolUtils;
import ru.ispras.microtesk.tools.adapter.trace.ConcreteEvent;
import ru.ispras.microtesk.tools.adapter.trace.EventListWriter;
import ru.ispras.microtesk.tools.adapter.trace.EventRecognizer;
import ru.ispras.microtesk.tools.adapter.trace.EventTemplateReader;

/**
 * The {@code Adapter} class adapts the program execution trace and writes it to file in JSON
 * format.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class Adapter {
  public static final String ADAPTER_OUTPUT_DIRECTORY = "adapter_output";

  private final String inputResourceDir;

  private final String outputAdapterDir;
  private final String outputAdapterFileSuffix;

  public Adapter(final Options options, final String inputResourceDir) {
    InvariantChecks.checkNotNull(options);
    InvariantChecks.checkNotNull(inputResourceDir);

    this.inputResourceDir = inputResourceDir;

    this.outputAdapterDir = ToolUtils.getOutDir(options, Adapter.ADAPTER_OUTPUT_DIRECTORY);
    this.outputAdapterFileSuffix = AdapterSettings.getAdapterEventSuffix(inputResourceDir);
  }

  public static boolean adapt(final Options options,
      final String inputTrace, final String inputResourceDir) {
    InvariantChecks.checkNotNull(options);
    InvariantChecks.checkNotNull(inputTrace);
    InvariantChecks.checkNotNull(inputResourceDir);

    Logger.setDebug(true);
    Logger.debug("Adapter running.");

    final Adapter adapter = new Adapter(options, inputResourceDir);

    final String outputTraceJsonFile = adapter.adapt(inputTrace);
    final Path adapterSettingsPath = Paths.get(outputTraceJsonFile);

    if (!Files.exists(adapterSettingsPath)) {
      return true;
    } else {
      return false;
    }
  }

  public String adapt(final String inputTrace) {
    // Trace events
    EventTemplateReader eventTemplateReader = new EventTemplateReader(inputResourceDir);
    EventRecognizer eventRecognizer = new EventRecognizer(eventTemplateReader.getEvents());
    List<ConcreteEvent> concreteEvents = eventRecognizer.getAllEvents(inputTrace);

    final String outputTraceJsonFile = this.getOutputAdapterFile(inputTrace);

    EventListWriter.saveEventsToJson(concreteEvents, outputTraceJsonFile);

    return outputTraceJsonFile;
  }

  private String getOutputAdapterFile(final String inputTrace) {
    return String.format("%s%s%s%s", this.getOutputAdapterDir(), File.separator,
        FileUtils.getShortFileNameNoExt(inputTrace), this.getOutputAdapterFileSuffix());
  }

  /**
   * Returns the adapter output directory.
   *
   * @return the adapter output directory
   */
  public String getOutputAdapterDir() {
    return outputAdapterDir;
  }

  /**
   * Returns the events json file suffix.
   *
   * @return the events json file suffix
   */
  public String getOutputAdapterFileSuffix() {
    return outputAdapterFileSuffix;
  }
}
