/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.adapter.trace;

import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import javax.json.Json;
import javax.json.JsonArray;
import javax.json.JsonObject;
import javax.json.JsonReader;
import javax.json.JsonReaderFactory;
import javax.json.JsonValue;

import ru.ispras.castle.util.FileUtils;
import ru.ispras.castle.util.Logger;

/**
 * The {@code EventElementPart} class reads RegEx from a JSON file to recognize the target part of
 * the event element.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class EventElementPart {
  private final Map<EventElementType, String> eventElementPartPatterns;

  public EventElementPart() {
    this.eventElementPartPatterns = new HashMap<>();
  }

  public EventElementPart(final Map<EventElementType, String> eventElementPartPatterns) {
    this.eventElementPartPatterns = eventElementPartPatterns;
  }

  public EventElementPart(final String inputFile) {
    this();

    final JsonReaderFactory factory = Json.createReaderFactory(null);
    try (final JsonReader jsonReader = factory.createReader(new FileReader(inputFile))) {
      final JsonArray jsonArray = jsonReader.readArray();
      Logger.debug("Read from json file: %s", FileUtils.getShortFileName(inputFile));

      for (JsonValue jsonValue : jsonArray) {
        final JsonObject jsonObject = jsonValue.asJsonObject();
        final Set<String> keySet = jsonObject.keySet();
        for (final String key : keySet) {
          this.addEventElementPartPattern(EventElementType.getType(key), jsonObject.getString(key));
        }
      }
    } catch (IOException e) {
      Logger.warning("Oracle tool: failed to read json file '%s': %s", inputFile, e.toString());
      e.printStackTrace();
    }
  }

  public void addEventElementPartPattern(final EventElementType eventElementType,
      final String regEx) {
    this.eventElementPartPatterns.put(eventElementType, regEx);

    Logger.debug("(class: %s) for event element part '%s' adds RegExps: '%s'",
        this.getClass().getSimpleName(), eventElementType.name(), regEx);
  }

  public Map<EventElementType, String> getElementPartPatterns(){
    return this.eventElementPartPatterns;
  }

  @Override
  public String toString() {
    final StringBuffer returnString = new StringBuffer("");
    returnString.append(String.format("[%s]: [", this.getClass().getSimpleName()));

    for (Entry<EventElementType, String> entry : eventElementPartPatterns.entrySet()) {
      returnString.append("{").append(entry.getKey().name()).append(" : \"")
          .append(entry.getValue()).append("\"}/n");
    }

    returnString.append("]");
    return returnString.toString();
  }
}
