/*
 * Copyright 2021-2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.adapter.trace;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.json.JsonArray;
import javax.json.JsonObject;
import javax.json.JsonReader;
import javax.json.JsonValue;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.utils.JsonConfig;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class EventListReader implements Events {
  private JsonReader jsonReader;

  // Current element
  private int index = 0;

  // Collection of recognized ({@code ConcreteEvent}) events in the trace
  public final List<ConcreteEvent> events = new ArrayList<>();;

  public EventListReader(final String fileName) {
    InvariantChecks.checkNotNull(fileName);
    jsonReader = JsonConfig.getJsonReader(fileName);

    this.readAll();
  }

  private void readAll() {
    JsonArray jsonArray = jsonReader.readArray();
    for (JsonValue jsonValue : jsonArray) {
      JsonObject jsonObject = jsonValue.asJsonObject();
      final Map<EventElementType, String> elements = new HashMap<>();
      EventType eventType = null;

      for (String key : jsonObject.keySet()) {
        if (EventType.EVENT_TYPE_IDENTIFIER.equals(key)) {
          eventType = EventType.valueOf(jsonObject.getString(key));
        } else {
          if (EventElementType.getType(key) == null) {
            Logger.error("Oracle tool: failed to get EventElementType for string: \"%s\"", key);
            throw new IllegalArgumentException(String.format("Oracle tool: failed to get EventElementType for string: \"%s\"", key));
          }

          elements.put(EventElementType.getType(key), jsonObject.getString(key));
        }
      }

      InvariantChecks.checkNotNull(eventType);
      events.add(new ConcreteEvent(eventType, elements));
    }
  }

  public void close() {
    JsonConfig.close(jsonReader);
  }

  public List<ConcreteEvent> getEvents() {
    return events;
  }

  @Override
  public boolean hasNext() {
    return events.size() > index;
  }

  @Override
  public ConcreteEvent readNext() {
    return events.get(index++);
  }

  @Override
  public ConcreteEvent getCurrentEvent() {
    return events.get(index);
  }
}