/*
 * Copyright 2021-2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.adapter.trace;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;

/**
 * The {@link EventRecognizer} class recognizes events according to templates.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class EventRecognizer {
  protected final List<EventTemplate> traceEvents;

  public EventRecognizer(final List<EventTemplate> events) {
    super();
    InvariantChecks.checkNotNull(events);

    this.traceEvents = events;
  }

  private EventTemplate getLineEvent (final String inputTrace) {
    for(EventTemplate eventTemplate : traceEvents) {
      if(eventTemplate.checkLineEventType(inputTrace)) {
        return eventTemplate;
      }
    }
    Logger.warning("[Class:%s] The event in the trace was not recognized:\n%s", this.getClass().getSimpleName(), inputTrace);

    // TODO:
    return null;
  }

  public List<ConcreteEvent> getAllEvents(final String inputTrace) {
    final List<ConcreteEvent> arrayElements = new ArrayList<>();
    try (final BufferedReader bufferedReader =
        new BufferedReader(new FileReader(new File(inputTrace)));) {
      String line;
      while ((line = bufferedReader.readLine()) != null) {
        final ConcreteEvent concreteEvent = this.getEvent(line);
        Logger.debug("%s", concreteEvent);

        if (!concreteEvent.getEventType().equals(EventType.INFO)) {
          arrayElements.add(concreteEvent);
        }
      }
    } catch (IOException e) {
      Logger.warning("Oracle tool: failed to read trace file '%s': %s", inputTrace, e.toString());
      e.printStackTrace();
    }

    return arrayElements;
  }

  public ConcreteEvent getEvent(final String eventLine) {
    final EventTemplate eventTemplate = this.getLineEvent(eventLine);

    if (eventTemplate != null) {
      Logger.debug("[%s] Type: %s, string: '%s'", this.getClass().getSimpleName(), eventTemplate.getEventType().name(),
          eventLine);

      final ConcreteEvent concreteEvent = parseLine(eventTemplate, eventLine);
      return concreteEvent;
    }
    // TODO: EventType.INFO branch

    return new ConcreteEvent(EventType.INFO, Map.of());
  }

  private ConcreteEvent parseLine(final EventTemplate event, final String line) {
    final List<EventElementPattern> events = event.getEventElements();
    final Map<EventElementType, String> eventElements = new HashMap<>();

    // The current position in the line
    int i = 0;

    for (final EventElementPattern eventElement : events) {
      // Get the pattern for current part of line
      Pattern p = Pattern.compile(eventElement.getRegEx());

      final String newLine = (String) line.subSequence(i, line.length());
      Matcher m = p.matcher(newLine);
      m.find(); // TODO: Exception
      final String partOfLine = m.group();
      i = i + m.end();
      //Logger.debug("'%s' - '%s' gets from event '%s'", partOfLine, eventElement, newLine);

      if (event.getElementsRegExp().containsKey(eventElement.getType())) {
        final String patternForElement = event.getElementsRegExp().get(eventElement.getType());
        p = Pattern.compile(patternForElement);
        m = p.matcher(partOfLine);

        m.find(); // TODO: Exception
        final String exactValue = m.group();
        // Logger.debug("'%s' - exact value, gets from element '%s'", exactValue, orderedElement);
        eventElements.put(eventElement.getType(), exactValue);
      }
    }

    return new ConcreteEvent(event.getEventType(), eventElements);
  }

  @Override
  public String toString() {
    final StringBuffer returnString = new StringBuffer("");
    returnString.append(String.format("[Class:%s]:", this.getClass().getSimpleName()));

    for (EventTemplate eventTemplate : traceEvents) {
      returnString.append("\n      {").append(eventTemplate.getEventType().name()).append(" : \"").append(eventTemplate.getEventPattern()).append("\"}");
    }

    return returnString.toString();
  }
}
