/*
 * Copyright 2021-2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.adapter.trace;

import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class EventTemplate {
  // "Pattern 1" + JUNCTION_ELEMENTS_PATTERN + "Pattern 2" ...
  private final static String JUNCTION_ELEMENT_PATTERN = "(.){1,}?"; // "(.){1,}?"

  // Map of elements and patterns for getting them from file
  private final List<EventElementPattern> eventElements;

  // The @EventType of this event
  private final EventType eventType;

  // The patterns for part of target elements. Example: "[TID 00755f]" -> "00755f"
  private Map<EventElementType, String> elementPartsRegExp;

  // The {@code pattern} field contains a pattern that allows you to check if the line is this event
  private final Pattern pattern;

  public EventTemplate(final EventType eventType, final List<EventElementPattern> eventElements,
      final Map<EventElementType, String> elementPartsRegExp) {
    InvariantChecks.checkNotNull(eventType);
    InvariantChecks.checkNotNull(eventElements);
    InvariantChecks.checkNotNull(elementPartsRegExp);

    this.eventType = eventType;
    this.eventElements = eventElements;
    this.pattern = this.createPattern(eventElements);
    this.elementPartsRegExp = elementPartsRegExp;
  }

  private Pattern createPattern(final List<EventElementPattern> eventElements) {
    final StringBuffer pattern = new StringBuffer("");

    boolean first = true;
    for (final EventElementPattern eventElement : eventElements) {
      if (!first) {
        pattern.append(JUNCTION_ELEMENT_PATTERN);
      }
      first = false;
      pattern.append(eventElement.getRegEx());
    }

    return Pattern.compile(pattern.toString());
  }

  public List<EventElementPattern> getEventElements() {
    return eventElements;
  }

  public EventType getEventType() {
    return eventType;
  }

  public Map<EventElementType, String> getElementsRegExp() {
    return elementPartsRegExp;
  }

  public Pattern getEventPattern() {
    return pattern;
  }

  /**
   * Returns the event type {@code EventType}.
   *
   * @param traceLine string from trace to check.
   *
   * @return the event type.
   */
  public boolean checkLineEventType(final String traceLine) {
    final Matcher matcher = pattern.matcher(traceLine);
    return matcher.find();
  }

  @Override
  public String toString() {
    final StringBuffer buffer = new StringBuffer();
    buffer.append(String.format("(%s) ", this.getClass().getSimpleName())).append("[Event : ")
        .append(eventType.name()).append("] Elements: [");

    boolean separator = false;
    for (EventElementPattern eventElement : eventElements) {
      buffer.append(separator ? " | " : "").append(eventElement);
      separator = true;
    }

    buffer.append("Pattern {\"").append(pattern).append("\"}");
    buffer.append("]");

    return buffer.toString();
  }
}
