/*
 * Copyright 2021-2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.adapter.trace;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.json.Json;
import javax.json.JsonArray;
import javax.json.JsonObject;
import javax.json.JsonReader;
import javax.json.JsonReaderFactory;
import javax.json.JsonValue;

import ru.ispras.castle.util.FileUtils;
import ru.ispras.castle.util.Logger;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class EventTemplateReader {
  // File suffix with templates for event recognition
  private static final String EVENTS_FILE_NAME   = "trace_adapter_event_template.json";

  // File suffix with templates for elements recognition
  private static final String ELEMENTS_FILE_NAME = "trace_adapter_element_template.json";

  //private Map<EventType, EventTemplate> events = new HashMap<>();
  private List<EventTemplate> eventsList = new ArrayList<>();

  EventElementPart eventElementPart;

  public EventTemplateReader(final String inputResourceDir) {
    final String traceEventElements = inputResourceDir + File.separator + ELEMENTS_FILE_NAME;
    final String traceEvents = inputResourceDir + File.separator + EVENTS_FILE_NAME;

    // Gets a significant element parts patterns for event
    eventElementPart = new EventElementPart(traceEventElements);

    this.readEventsFromFile(traceEvents);
  }
  
  public List<EventTemplate> getEvents() {
    return eventsList;
  }

  private List<EventTemplate> readEventsFromFile(final String inputTemplateFile) {
    final JsonReaderFactory factory = Json.createReaderFactory(null);
    try (final JsonReader jsonReader = factory.createReader(new FileReader(inputTemplateFile));) {
      final JsonArray jsonArray = jsonReader.readArray();
      Logger.debug("Read from json file: %s", FileUtils.getShortFileName(inputTemplateFile));

      for (JsonValue jsonValue : jsonArray) {
        final JsonObject jsonObject = jsonValue.asJsonObject();
        final Set<String> keySet = jsonObject.keySet();

        final List<EventElementPattern> eventElements = new ArrayList<>();
        EventType type = null;

        for (String key : keySet) {
          final String value = jsonObject.getString(key);
          //Logger.debug("   '%s' : '%s'", key, value);

          if (EventType.EVENT_TYPE_IDENTIFIER.equals(key)) {
            type = EventType.valueOf(value);
          } else {
            if (EventElementType.getType(key) == null) {
              final String className = EventElementType.class.getSimpleName();
              Logger.error("Oracle tool: failed to get '%s' for string: \"%s\"", className, key);
              throw new IllegalArgumentException(String.format("Oracle tool: failed to get '%s' for string: \"%s\"", className, key));
            }
            EventElementPattern eventElement = new EventElementPattern(EventElementType.getType(key), value);
            Logger.debug("   '%s' ", eventElement);
            eventElements.add(eventElement);
          }
        }
        Map<EventElementType, String> eventElementsPartPatterns = this.getElementPartPattern(eventElements);

     //   this.events.put(type, new EventTemplate(type, eventElements, eventElementsPartPatterns));
      //  Logger.debug("Adds the %s", this.events.get(type));

        this.eventsList.add(new EventTemplate(type, eventElements, eventElementsPartPatterns));
        Logger.debug("Adds the %s", eventsList.get(eventsList.size() - 1));
      }
    } catch (IOException e) {
      Logger.warning("Oracle tool: failed to read json file '%s': %s", inputTemplateFile, e.toString());
      e.printStackTrace();
    }

    return this.eventsList;
  }

  private Map<EventElementType, String> getElementPartPattern(final List<EventElementPattern> eventElements) {
    // Adds patterns for target event elements

    final Map<EventElementType, String> elementsRegExp = new HashMap<>();
    Map<EventElementType, String> eventElementPartPatterns = eventElementPart.getElementPartPatterns();

    for (EventElementPattern element : eventElements) {
      EventElementType elementType = element.getType();

      if (eventElementPartPatterns.containsKey(elementType)) {
      elementsRegExp.put(elementType, eventElementPartPatterns.get(elementType));
      Logger.debug("(class: %s) For event to '%s' field adds RegExps: '%s'.",
          this.getClass().getSimpleName(), elementType.name(),
          eventElementPart.getElementPartPatterns().get(elementType));
      }
    }

    return elementsRegExp;
  }
}
