/*
 * Copyright 2021-2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.oracle;

import java.io.File;
import java.util.List;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.model.Model;
import ru.ispras.microtesk.options.Option;
import ru.ispras.microtesk.options.Options;
import ru.ispras.microtesk.tools.ToolUtils;
import ru.ispras.microtesk.tools.adapter.Adapter;
import ru.ispras.microtesk.tools.adapter.trace.EventListReader;
import ru.ispras.microtesk.tools.oracle.internal.SourceFiles;
import ru.ispras.microtesk.tools.oracle.matcher.EventList;
import ru.ispras.microtesk.tools.simulator.IrProgram;
import ru.ispras.microtesk.tools.simulator.ProgramImage;
import ru.ispras.microtesk.tools.simulator.ProgramReader;
import ru.ispras.microtesk.tools.simulator.ProgramReaderFactory;
import ru.ispras.microtesk.tools.simulator.ProgramSimulator;
import ru.ispras.microtesk.tools.simulator.Simulator;
import ru.ispras.microtesk.tools.simulator.SimulatorSettings;
import ru.ispras.microtesk.tools.simulator.memory.IrMemory;
import ru.ispras.microtesk.tools.simulator.memory.MemoryFactory;
import ru.ispras.microtesk.tools.simulator.memory.MemoryFiles;
import ru.ispras.microtesk.tools.simulator.settings.SimulatorParameters;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class Oracle {
  public static final String ORACLE_OUTPUT_DIRECTORY = "oracle_output";

  // Model needed to decode {@link ProgramReader} and simulate {@link ProgramSimulator } the program
  private final Model model;
  private final Options options;
  private final String inputResourceDir;

  private final String outputOracleDir;
  private final String outputOracleLogFile;

  public static boolean predict(final Options options, final String modelName,
      final String inputProgram, final String inputTrace, final String inputResourceDir) {
    InvariantChecks.checkNotNull(options);
    InvariantChecks.checkNotNull(modelName);
    InvariantChecks.checkNotNull(inputProgram);
    InvariantChecks.checkNotNull(inputTrace);
    InvariantChecks.checkNotNull(inputResourceDir);

    // TODO:
    Logger.setDebug(true);

    Oracle oracle = new Oracle(options, modelName, inputResourceDir);

    boolean allPrograms = options.getValueAsBoolean(Option.RUN_ALL_PROGRAMS);
    List<SourceFiles> comparedFiles =
        SourceFiles.getComparedFilesGroup(inputProgram, inputTrace, allPrograms);

    return oracle.run(comparedFiles);
  }

  private Oracle(final Options options, final String modelName, final String inputResourceDir) {
    InvariantChecks.checkNotNull(options);
    InvariantChecks.checkNotNull(modelName);
    InvariantChecks.checkNotNull(inputResourceDir);

    this.options = options;
    this.model = ToolUtils.loadModel(modelName);
    this.inputResourceDir = inputResourceDir;

    this.outputOracleDir = ToolUtils.getOutDir(options, Oracle.ORACLE_OUTPUT_DIRECTORY);
    this.outputOracleLogFile = OracleSettings.getOracleLogFileName(inputResourceDir);
  }

  private boolean run(List<SourceFiles> comparedFiles) {
    final String logFileFullName = String.format("%s%s%s", this.getOutputOracleDir(),
        File.separator, this.getOutputOracleLogFile());
    Statistic statistic = new Statistic(logFileFullName, comparedFiles.size() == 1);
    boolean proceed = true;

    int traceSuccess = 0;
    int traceFail = 0;

    for (SourceFiles oracleFiles : comparedFiles) {
      final boolean correctness = this.simulateProgramm(oracleFiles);
      if (correctness) {
        traceSuccess++;
      } else {
        proceed = false;
        traceFail++;
      }

      String tempLog = statistic.writeSimpleResult(oracleFiles.inputProgram, correctness);
      if (correctness) {
        Logger.debug(tempLog);
      } else {
        Logger.warning(tempLog);
      }

      proceed &= correctness;
    }

    if (comparedFiles.size() > 1) {
      statistic.writeResult(traceSuccess, traceFail);
    }
    statistic.close();

    return proceed;
  }

  private boolean simulateProgramm(final SourceFiles oracleFiles) {
    Logger.debug("Program file: %s, trace file: %s", oracleFiles.getInputProgram(),
        oracleFiles.getInputTrace());

    // Recognizing events in the trace
    final String inputTrace = oracleFiles.getInputTrace();
    final Adapter adapter = new Adapter(options, inputResourceDir);
    final String outputTraceJsonFile = adapter.adapt(inputTrace);

    // Reading a list of events from a json file
    final EventListReader eventListReader = new EventListReader(outputTraceJsonFile);
    final EventList eventReader = new EventList(eventListReader.getEvents());
    eventListReader.close();

    // Files & options for simulator
    final SimulatorSettings simulatorSettings =
        new SimulatorSettings(inputResourceDir, oracleFiles.getInputProgram(), options);
    final MemoryFiles memoryFiles = simulatorSettings.getMemoryFiles();

    // Options for simulator
    final SimulatorParameters simulatorParameters = simulatorSettings.getSimulatorParameters();

    // Logger.debug
    SimulatorParameters.printParameters(Simulator.class.getSimpleName(), simulatorParameters);

    // Disassemble -> Program
    ProgramReader programReader =
        ProgramReaderFactory.getBinaryProgramReader(model, options, simulatorSettings.getProgramFileName());
    final IrProgram simulatorProgram = programReader.getProgram();

    // Memory
    final IrMemory simulatorMemory = MemoryFactory.newMemory(model.getName(), memoryFiles);
    Logger.debug("Simulator memory: %s", simulatorMemory.toString());

    // Simulator for program
    ProgramImage programImage = new ProgramImage(model, simulatorProgram, simulatorMemory, simulatorParameters);
    ProgramSimulator programSimulator = new ProgramSimulator(programImage);

    // Predicts program execution and compares to the trace events
    OracleSimulator oracleSimulator =
        new OracleSimulator(eventReader, programSimulator, inputResourceDir);

    return oracleSimulator.run();
  }

  /**
   * Returns the oracle output directory.
   *
   * @return the oracle output directory
   */
  public String getOutputOracleDir() {
    return outputOracleDir;
  }

  /**
   * Returns the oracle log file name.
   *
   * @return the oracle log file name
   */
  public String getOutputOracleLogFile() {
    return outputOracleLogFile;
  }
}