/*
 * Copyright 2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.oracle;

import java.io.File;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.tools.oracle.internal.Registers;
import ru.ispras.microtesk.tools.oracle.matcher.EventList;
import ru.ispras.microtesk.tools.oracle.matcher.OracleEventMatcher;
import ru.ispras.microtesk.tools.oracle.matcher.OracleEventMatcherFull;
import ru.ispras.microtesk.tools.oracle.matcher.OracleEventMatcherNotFull;
import ru.ispras.microtesk.tools.simulator.ProgramSimulator;

/**
 * The {@link OracleSimulator} class runs the realization of {@link OracleEventMatcher}.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class OracleSimulator {

  private final EventList eventReader;
  private final ProgramSimulator programSimulator;

  //
  private final Registers registers;

  /**
   * Creates an oracle that compares the output of the model with the control trace
   *
   * @param eventReader the trace events.
   * @param programSimulator Gives the state of the model for comparison.
   * @param inputResourceDirTemp the path to resources.
   *
   * @throws IllegalArgumentException if any of the parameters equals {@code null}.
   */
  public OracleSimulator(final EventList eventReader, final ProgramSimulator programSimulator,
      final String inputResourceDirTemp) {
    InvariantChecks.checkNotNull(eventReader);
    InvariantChecks.checkNotNull(programSimulator);
    InvariantChecks.checkNotNull(inputResourceDirTemp);

    this.eventReader = eventReader;
    this.programSimulator = programSimulator;
    this.registers =
        new Registers(inputResourceDirTemp + File.separator + Registers.REGISTERS_DESCRIPTION);
  }

  public boolean run() {
    OracleEventMatcherNotFull matcher =
        new OracleEventMatcherNotFull(eventReader, programSimulator, registers);

    boolean proceed = matcher.hasNext() ? true : false;

    boolean matchResult = true;

    while (proceed && matcher.hasNext()) {
      int eventCheckIndex = matcher.getEventIndex();
      proceed = matcher.checkEvent();

      if(!proceed) {
        // Logger.debug("________ Not match");
        while(!proceed && matcher.hasNext()) {
          // Logger.debug("________ Not match cycle");
          matcher.setEventIndex(eventCheckIndex);
          matcher.nextProgramStep();
          proceed = matcher.checkEvent();
        }
      }
      if (matcher.hasNext()) {
        matcher.nextEvent();
      }

      matchResult &= proceed;
    }

    if(matchResult && matcher.getSkippedInstructions().size() > 0) {
      Logger.debug("Program instructions that were skipped in the trace:");
      for(String skippedInstruction: matcher.getSkippedInstructions()) {
        Logger.debug("%s", skippedInstruction);
      }
    }

    if(!matcher.isAllTraceEventsMatched()) {
      Logger.error("[OracleSimulator] Not all events from the trace are matched.");
    }

    return matchResult && matcher.isAllTraceEventsMatched();
  }

  public boolean runFull() {
    OracleEventMatcher matcher =
        new OracleEventMatcherFull(eventReader, programSimulator, registers);

    boolean proceed = true;
    while (proceed) {
      proceed = matcher.checkEvent();

      if(matcher.hasNext()) {
        matcher.nextEvent();
      } else {
        break;
      }
    }

    return proceed;
  }
}
