/*
 * Copyright 2022-2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.oracle.internal;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import ru.ispras.fortress.util.InvariantChecks;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class RegisterMode implements Item {
  private final String nameRegex;
  private final String memoryName;
  private final boolean frameDependence;
  private final int startIndex;
  private final int endIndex;

  public RegisterMode(final String nameRegex, final String memoryName, final boolean frameDependence, final int startIndex,
      final int endIndex) {
    InvariantChecks.checkNotNull(nameRegex);
    InvariantChecks.checkNotNull(memoryName);

    this.nameRegex = nameRegex;
    this.memoryName = memoryName;
    this.frameDependence = frameDependence;
    this.startIndex = startIndex;
    this.endIndex = endIndex;
  }

  @Override
  public String getItemName() {
    return memoryName;
  }

  @Override
  public int getIndex(final String register) {
    Pattern pattern = Pattern.compile(nameRegex);
    Matcher matcher = pattern.matcher(register);

    if (matcher.find()) {
      if (endIndex - startIndex > 0) {
        String index = matcher.group(1);
        int regIndex = Integer.decode(index).intValue();

        if (regIndex > endIndex) {
          throw new IllegalArgumentException();
        } else {
          return regIndex;
        }
      } else {
        return 0;
      }
    } else {
      throw new IllegalArgumentException();
    }
  }

  @Override
  public boolean check(String item) {
    Pattern pattern = Pattern.compile(nameRegex);
    Matcher matcher = pattern.matcher(item);

    if (matcher.find()) {
      return true;
    } else {
      return false;
    }
  }

  @Override
  public String toString() {
    return String.format("Regex for register: \"%s\", memory name: \"%s\", frame dependence: \"%s\", index range[%s, %s]",
        nameRegex, memoryName, frameDependence, startIndex, endIndex);
  }

  @Override
  public int getEndIndex() {
    return endIndex;
  }

  @Override
  public boolean getFrameDependence() {
    return frameDependence;
  }

  @Override
  public String getShortDescription(int index) {
    if (endIndex > 0) {
      return String.format("%s[%d]", memoryName, index);
    } else {
      return String.format("%s", memoryName);
    }
  }
}
