/*
 * Copyright 2022-2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.oracle.internal;

import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.NoSuchElementException;

import javax.json.Json;
import javax.json.JsonArray;
import javax.json.JsonObject;
import javax.json.JsonReader;
import javax.json.JsonReaderFactory;
import javax.json.JsonValue;

import ru.ispras.castle.util.Logger;
import ru.ispras.microtesk.utils.JsonConfig;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class Registers {
  public static final String REGISTERS_DESCRIPTION = "oracle_trace_registers.json";

  private static final String REGISTER_REGEX   = "RegisterRegEx";
  private static final String MEMORY_DEVICE    = "MemoryDevice";
  private static final String FRAME_DEPENDENCE = "FrameDependence";
  private static final String INDEX_START      = "IndexStart";
  private static final String INDEX_END        = "IndexEnd";

  // Contains the path to information for trace recognition
  private final String resourcePath;

  // Contains the oracle registers templates
  private final List<Item> oracleItems = new ArrayList<>();

  /**
   * @return the path to information about the oracle internal items
   */
  public String getResourcePath() {
    return resourcePath;
  }

  public Registers(final String resourcePath) {
    this.resourcePath = resourcePath;

    this.createRegistersList();
  }

  private void createRegistersList() {
    final String fileName = resourcePath;
    Logger.debug("The names of the oracle registers are loaded from the file: '%s'", fileName);

    JsonReaderFactory factory = Json.createReaderFactory(JsonConfig.getConfig());
    try (final JsonReader jsonReader = factory.createReader(new FileReader(fileName));) {
      final JsonArray jsonArray = jsonReader.readArray();
      for (JsonValue jsonValue : jsonArray) {
        final JsonObject jsonObject = jsonValue.asJsonObject();

        final String registerRegex = jsonObject.getString(REGISTER_REGEX);
        final String memoryName = jsonObject.getString(MEMORY_DEVICE);
        final boolean frameDependence = jsonObject.getBoolean(FRAME_DEPENDENCE);
        final Integer indexStart = Integer.parseInt(jsonObject.getString(INDEX_START));
        final Integer indexEnd = Integer.parseInt(jsonObject.getString(INDEX_END));

        oracleItems.add(new RegisterMode(registerRegex, memoryName, frameDependence, indexStart, indexEnd));
      }
    } catch (IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  public Item recognizeItem(final String item) {
    for (final Item oracleItem : this.oracleItems) {
      if (oracleItem.check(item)) {
        return oracleItem;
      }
    }

    throw new NoSuchElementException(String.format("Trace item: '%s' cannot be recognized.", item));
  }
}
