/*
 * Copyright 2022-2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.oracle.internal;

import java.io.File;
import java.io.FilenameFilter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.stream.Collectors;

import ru.ispras.castle.util.FileUtils;
import ru.ispras.castle.util.Logger;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class SourceFiles {
  public final String inputProgram;
  public final String inputTrace;

  public SourceFiles(final String inputProgram, final String inputTrace) {
    this.inputProgram = inputProgram;
    this.inputTrace = inputTrace;
  }

  public String getInputProgram() {
    return inputProgram;
  }

  public String getInputTrace() {
    return inputTrace;
  }

  public static List<SourceFiles> getComparedFilesGroup(final String program, final String trace,
      final boolean all) {
    if (all) {
      return SourceFiles.getAllComparedFiles(program, trace);
    } else {
      List<SourceFiles> comparedFiles = new ArrayList<>();
      comparedFiles.add(new SourceFiles(program, trace));
      return comparedFiles;
    }
  }

  /**
   * @param inputProgram name of program for oracle
   * @param inputTrace name of trace for oracle
   * 
   * @return the list of grouped input files
   */
  public static List<SourceFiles> getAllComparedFiles(final String inputProgram,
      final String inputTrace) {
    List<SourceFiles> comparedFiles = new ArrayList<>();
    final Map<String, String> mapProgram = getInputFilesMap(inputProgram);
    final Map<String, String> mapTrace = getInputFilesMap(inputTrace);

    for (Entry<String, String> entry : mapProgram.entrySet()) {
      if (mapTrace.containsKey(entry.getKey())) {
        comparedFiles.add(new SourceFiles(entry.getValue(), mapTrace.get(entry.getKey())));
      }
    }

    if (comparedFiles.size() == 0) {
      Logger.error("No input file pairs exist for files: %s, %s", inputProgram, inputTrace);
    }

    return comparedFiles;
  }

  // Returns Map<key, path to file> of input files
  private static Map<String, String> getInputFilesMap(final String inputFile) {
    final File inputSourceFile = new File(inputFile);
    if (!inputSourceFile.exists()) {
      Logger.error("Input file: %s - does not exist", inputFile);
      return null;
    }

    String inputSourceFileExtension = FileUtils.getFileExtension(inputSourceFile.getName());
    String inputSourceFileDir = inputSourceFile.getParent();

    Logger.debug("File directory: %s, extension: %s", inputSourceFileDir, inputSourceFileExtension);

    File inputFileDir = new File(inputSourceFileDir);

    File[] inputListFile = inputFileDir.listFiles(new FilenameFilter() {
      public boolean accept(File dir, String name) {
        return name.matches(".+\\" + inputSourceFileExtension);
      }
    });

    Logger.debug("Input files: %s", Arrays.toString(inputListFile));

    Map<String, String> mapProgram = Arrays.asList(inputListFile).stream().collect(
        Collectors.toMap(h -> FileUtils.getShortFileNameNoExt(h.getName()), File::getPath));

    Logger.debug("Created map: %s", mapProgram);

    return mapProgram;
  }
}
