/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.oracle.matcher;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import ru.ispras.castle.util.Logger;
import ru.ispras.microtesk.tools.adapter.trace.ConcreteEvent;
import ru.ispras.microtesk.tools.adapter.trace.EventElementType;
import ru.ispras.microtesk.tools.adapter.trace.EventType;

/**
 * The {@link EventList} class returns trace events one by one.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class EventTransformer {
  private final Map<String, ConcreteEvent> mapThreadPull= new HashMap<>();

  public EventTransformer() {
  }

  /*
    Normalizes the list of trace events. For example: EventA.instruction: 'add ->' +
    EventB.instruction: 'binop2 v1, v2' => EventAB.instruction: 'add v1, v2'
   */
  public List<ConcreteEvent> normalizeEvents(final List<ConcreteEvent> events) {
    final List<ConcreteEvent> eventsList = new ArrayList<>();
    for (ConcreteEvent concreteEvent: events) {
      // TODO: We need a special file to describe the normalization rules.
      if (EventType.INSTRUCTION.equals(concreteEvent.getEventType())) {
        if (pandaTransformationNeeded(concreteEvent.getElement(EventElementType.INSTRUCTION))) {
          mapThreadPull.put(concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER), concreteEvent);
        } else {
          final String eventThread = concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER);
          if (mapThreadPull.containsKey(eventThread)) {
            eventsList.add(pandaInstructionTranform(mapThreadPull.get(eventThread), concreteEvent));
            mapThreadPull.remove(eventThread);
          } else {
            eventsList.add(concreteEvent);
          }
        }
      } else {
        eventsList.add(concreteEvent);
      }
    }

    return Collections.unmodifiableList(eventsList);
  }

  // TODO:  We need a special file to describe the normalization rules.
  private boolean pandaTransformationNeeded(final String instruction) {
    final Pattern p = Pattern.compile("(.*)( ->)");
    final Matcher m = p.matcher(instruction);
    return m.matches();
  }

  // TODO:  We need a special file to describe the normalization rules.
  private ConcreteEvent pandaInstructionTranform(final ConcreteEvent first,
      final ConcreteEvent second) {
    final String firstInstruction = first.getElement(EventElementType.INSTRUCTION);
    final String secondInstruction = second.getElement(EventElementType.INSTRUCTION);

    Pattern p = Pattern.compile("(.*)( ->)");
    Matcher m = p.matcher(firstInstruction);
    m.find();
    final String firstPart = m.group(1);

    //p = Pattern.compile("([a-z0-9]{2,}+)(.*)");
    p = Pattern.compile("([a-z0-9]*, )*[a-z0-9]+$");
    m = p.matcher(secondInstruction);
    m.find();
    final String secondPart = m.group(0);

    final String tranformSyntax = firstPart + " " + secondPart;
    Logger.debug("Merge events of instructions '%s' + '%s' => '%s'", firstInstruction,
        secondInstruction, tranformSyntax);

    final Map<EventElementType, String> eventElements = second.getEventElements();
    eventElements.put(EventElementType.INSTRUCTION, tranformSyntax);
    return new ConcreteEvent(second.getEventType(), eventElements);
  }

}
