/*
 * Copyright 2021-2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.oracle.matcher;

import java.util.HashMap;
import java.util.Map;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.tools.adapter.trace.ConcreteEvent;
import ru.ispras.microtesk.tools.adapter.trace.EventElementType;
import ru.ispras.microtesk.tools.adapter.trace.EventType;
import ru.ispras.microtesk.tools.oracle.internal.Item;
import ru.ispras.microtesk.tools.oracle.internal.Registers;
import ru.ispras.microtesk.tools.simulator.ProgramSimulator;

/**
 * The {@link OracleEventMatcherFull} class runs the program on {@link ProgramSimulator} step by step and
 * compares the result with the events from the control trace
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class OracleEventMatcherFull implements OracleEventMatcher {
  protected final EventList eventReader;
  protected final ProgramSimulator programSimulator;

  protected final Registers registers;
  protected final Map<String, Long> objIdentifiers = new HashMap<>();
  protected Map<String, Long> objIdentifiersNew;

  /**
   * Creates the matcher that compares the output of the model with the control trace
   *
   * @param eventReader the trace events.
   * @param programSimulator Gives the state of the model for comparison.
   * @param registers the simulator registers.
   *
   * @throws IllegalArgumentException if any of the parameters equals {@code null}.
   */
  public OracleEventMatcherFull(final EventList eventReader,
      final ProgramSimulator programSimulator, final Registers registers) {
    InvariantChecks.checkNotNull(eventReader);
    InvariantChecks.checkNotNull(programSimulator);
    InvariantChecks.checkNotNull(registers);

    this.eventReader = eventReader;
    this.programSimulator = programSimulator;
    this.registers = registers;
  }

  @Override
  public void nextEvent() {
    eventReader.readNext();
  }

  protected ConcreteEvent getCurrentEvent() {
    return eventReader.getCurrentEvent();
  }

  @Override
  public boolean checkEvent() {
    final EventType eventType = getCurrentEvent().getEventType();
    switch (eventType) {
      case MEMORY:
        objIdentifiersNew = new HashMap<>();
        boolean checkReg = checkRegister();
        if(checkReg) {
          objIdentifiers.putAll(objIdentifiersNew);
        }
        return checkReg;
      case INSTRUCTION:
        this.checkProgramCounter();
        return checkInstruction();
      case FRAME_ENTRY:
        // TODO: Always passes check
        Logger.warning("Oracle tool: '%s'", this.getCurrentEvent());
        return true;
      case FRAME_EXIT:
        // TODO: Always passes check
        Logger.warning("Oracle tool: '%s'", this.getCurrentEvent());
        return true;
      default:
        Logger.error("Oracle tool: unknown event '%s'", this.getCurrentEvent());
        return false;
    }
  }

  protected boolean checkInstruction() {
    final ConcreteEvent concreteEvent = eventReader.getCurrentEvent();

    final String instruction = concreteEvent.getElement(EventElementType.INSTRUCTION);

    if (!programSimulator.hasNext()) {
      return false;
    }

    final String simulatorInstruction = programSimulator.nextInstructionText();


    if (instruction.equals(simulatorInstruction)) {
      OracleEventMatcherFull.printCheckInstruction(OracleEventMatcher.SUCCESSFUL_MATCH, instruction,
          simulatorInstruction);
      programSimulator.nextStep();
      return true;
    } else {
      OracleEventMatcherFull.printCheckInstruction(OracleEventMatcher.FAILURE_MATCH, instruction,
          simulatorInstruction);
      return false;
    }
  }

  private static void printCheckInstruction(final String checkStatus, final String instruction,
      final String simulatorInstruction) {
    Logger.debug("[%s] Compare Instruction: [%s], (trace : model) -> '%s' : '%s'",
        OracleEventMatcherFull.class.getSimpleName(), checkStatus, simulatorInstruction,
        instruction);
  }

  protected boolean checkProgramCounter() {
    final ConcreteEvent concreteEvent = eventReader.getCurrentEvent();

    final long simulatorPc = programSimulator.getCurrentPc();
    final long eventPc = Long.parseUnsignedLong(concreteEvent.getElement(EventElementType.PC), 16);

    if (simulatorPc == eventPc) {
      OracleEventMatcherFull.printCheckProgramCounter(SUCCESSFUL_MATCH, simulatorPc, eventPc);
      return true;
    } else {
      OracleEventMatcherFull.printCheckProgramCounter(FAILURE_MATCH, simulatorPc, eventPc);
      return false;
    }
  }

  private static void printCheckProgramCounter(final String checkStatus, final long simulatorPc,
      final long eventPc) {
    Logger.debug("[%s] Compare program counter (PC): [%s] (trace : model) -> '%s' : '%s'",
        OracleEventMatcherFull.class.getSimpleName(), checkStatus, Long.toHexString(eventPc),
        Long.toHexString(simulatorPc));
  }

  protected boolean checkRegister() {
    final ConcreteEvent concreteEvent = eventReader.getCurrentEvent();
    final String eventRegName = concreteEvent.getElement(EventElementType.REGISTER);
    final String eventRegValue = concreteEvent.getElement(EventElementType.REGISTER_VALUE);

    Item registerMode = registers.recognizeItem(eventRegName);
    final int index = registerMode.getIndex(eventRegName);

    final long modelRegValue = programSimulator.getRegisterValue(registerMode, index);
    final long traceValue = Long.parseUnsignedLong(eventRegValue, 16);

    // TODO:
    final String eventRegType = concreteEvent.getElement(EventElementType.REGISTER_TYPE);

    // Register contains system dependent unique value (example: object id).
    if (eventRegType != null && eventRegType.equals("obj")) {
      // TODO: obj: add unique value to setting for tool.

      String tempStatusEventRegValue;
      if (objIdentifiers.containsKey(eventRegValue)) {
        Long objValue = objIdentifiers.get(eventRegValue);
        if (!objValue.equals(modelRegValue)) {
          tempStatusEventRegValue = "No";
        } else {
          tempStatusEventRegValue = "Yes";
        }
      } else {
        objIdentifiersNew.put(eventRegValue, Long.valueOf(modelRegValue));
        tempStatusEventRegValue = "New";
      }

      OracleEventMatcherFull.printCheckRegister(
          String.format("Object matching: %s", tempStatusEventRegValue), eventRegName,
          eventRegValue, registerMode, index, modelRegValue);
      return true;
    } else if ( modelRegValue == traceValue) {
      // Register contains primitive
      OracleEventMatcherFull.printCheckRegister(OracleEventMatcher.SUCCESSFUL_MATCH, eventRegName, eventRegValue,
          registerMode, index, modelRegValue);
      return true;
    } else {
      OracleEventMatcherFull.printCheckRegister(OracleEventMatcher.FAILURE_MATCH, eventRegName, eventRegValue,
          registerMode, index, modelRegValue);
      return false;
    }
  }

  private static void printCheckRegister(final String checkStatus, final String eventRegName,
      final String eventRegValue, final Item registerMode, final int index,
      final long modelRegValue) {
    Logger.debug("[%s] Compare Registers: [%s], (trace : model) -> '%s' = 0x%s : '%s' = 0x%s",
        OracleEventMatcherFull.class.getSimpleName(), checkStatus, eventRegName, eventRegValue,
        registerMode.getShortDescription(index), Long.toHexString(modelRegValue));
  }

  @Override
  public boolean hasNext() {
    return eventReader.hasNext() && programSimulator.hasNext();
  }

  @Override
  public boolean isAllTraceEventsMatched() {
    return !eventReader.hasNext();
  }
}
