/*
 * Copyright 2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.oracle.matcher;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import ru.ispras.castle.util.Logger;
import ru.ispras.microtesk.tools.adapter.trace.ConcreteEvent;
import ru.ispras.microtesk.tools.adapter.trace.EventType;
import ru.ispras.microtesk.tools.oracle.internal.Registers;
import ru.ispras.microtesk.tools.simulator.ProgramSimulator;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class OracleEventMatcherNotFull extends OracleEventMatcherFull{

  private List<String> skippedInstructions = new ArrayList<>();

  /**
   * Creates the matcher that compares the output of the model with the control trace
   *
   * @param eventReader the trace events.
   * @param programSimulator Gives the state of the model for comparison.
   * @param registers the simulator registers.
   *
   * @throws IllegalArgumentException if any of the parameters equals {@code null}.
   */
  public OracleEventMatcherNotFull  (final EventList eventReader,
      final ProgramSimulator programSimulator, final Registers registers) {
    super(eventReader, programSimulator, registers);
  }

  @Override
  public boolean checkEvent() {
    final EventType eventType = getCurrentEvent().getEventType();
    switch (eventType) {
      case MEMORY:
        return checkRegistersBlock();
      case INSTRUCTION:
        this.checkProgramCounter();
        return checkInstructionAndSave();
      case FRAME_ENTRY:
        // TODO: Always passes check
        Logger.warning("Oracle tool: '%s'", this.getCurrentEvent());
        return true;
      case FRAME_EXIT:
        // TODO: Always passes check
        Logger.warning("Oracle tool: '%s'", this.getCurrentEvent());
        return true;
      default:
        Logger.error("Oracle tool: unknown event '%s'", this.getCurrentEvent());
        return false;
    }
  }

  public List<String> getSkippedInstructions() {
    return skippedInstructions;
  }

  private boolean checkInstructionAndSave() {
    boolean checkStatus = checkInstruction();

    return checkStatus;
  }

  private boolean checkRegistersBlock() {
    boolean checkRegistersFlag = checkRegister();

    ConcreteEvent nextEvent = eventReader.getNextEvent();
    EventType eventType = nextEvent.getEventType();

    objIdentifiersNew = new HashMap<>();
    while (eventType.equals(EventType.MEMORY)) {
      this.nextEvent();
      checkRegistersFlag = checkRegistersFlag & checkRegister();
      nextEvent = eventReader.getNextEvent();
      if (nextEvent == null) {
        Logger.debug("[%s] There are no more events in the trace.",
            OracleEventMatcherNotFull.class.getSimpleName());
        return false;
      }
      eventType = nextEvent.getEventType();
    }

    if(checkRegistersFlag) {
      objIdentifiers.putAll(objIdentifiersNew);
    }

    return checkRegistersFlag;
  }

  public void nextProgramStep() {
    skippedInstructions.add(String.format("PC[%s]: %s", programSimulator.getCurrentPc(), programSimulator.nextInstructionText()));

    //Logger.debug("%s", skippedInstructions.get(0));
    programSimulator.nextStep();
  }

  public int getEventIndex() {
    return eventReader.getIndex();
  }

  public void setEventIndex(int a) {
    eventReader.setIndex(a);
  }
}
