/*
 * Copyright 2021 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.simulator;

import java.util.ArrayList;
import java.util.List;

import ru.ispras.castle.util.FileUtils;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.model.IsaPrimitive;
import ru.ispras.microtesk.model.Model;
import ru.ispras.microtesk.options.Options;
import ru.ispras.microtesk.tools.Disassembler;
import ru.ispras.microtesk.tools.Disassembler.Output;
import ru.ispras.microtesk.tools.Disassembler.OutputFactory;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public abstract class BinaryProgramReader implements ProgramReader {
  protected final Model model;
  protected final Options options;
  protected final String programFileName;
  protected final String programFileNameNoExt;

  BinaryProgramReader(final Model model, final Options options, final String programFileName,
      final String programFileNameNoExt) {
    InvariantChecks.checkNotNull(model);
    InvariantChecks.checkNotNull(options);
    InvariantChecks.checkNotNull(programFileName);

    this.model = model;
    this.options = options;
    this.programFileName = programFileName;
    this.programFileNameNoExt = programFileNameNoExt;
  }

  BinaryProgramReader(final Model model, final Options options, final String programFileName) {
    this(model, options, programFileName, FileUtils.getShortFileNameNoExt(programFileName));
  }

  private static interface OutputProgram extends OutputFactory {
    Output getOutput();
  }

  private static final class OutputImpl implements Output {
    private final List<IsaPrimitive> program;

    private OutputImpl() {
      program = new ArrayList<IsaPrimitive>();
    }

    @Override
    public void add(final IsaPrimitive primitive) {
      program.add(primitive);
    }

    @Override
    public void close() {}
  }

  protected List<IsaPrimitive> disassembleProgram() {
    final OutputProgram outputFactory = new OutputProgram() {
      private Output outputProgram;

      @Override
      public Output createOutput(final Model model) {
        outputProgram = new OutputImpl();
        return outputProgram;
      }

      @Override
      public Output getOutput() {
        return outputProgram;
      }
    };

    Disassembler.disassemble(options, model.getName(), programFileName, outputFactory);

    final OutputImpl outputProgram = (OutputImpl) outputFactory.getOutput();
    return outputProgram.program;
  }
}
