/*
 * Copyright 2021 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.simulator;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.List;

import net.fornwall.jelf.ElfFile;
import net.fornwall.jelf.ElfSection;
import ru.ispras.castle.util.FileUtils;
import ru.ispras.castle.util.Logger;
import ru.ispras.microtesk.model.IsaPrimitive;
import ru.ispras.microtesk.model.Model;
import ru.ispras.microtesk.options.Options;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public final class ElfProgramReader extends BinaryProgramReader {
  protected final String fileName;

  public static final String FILE_EXTENSION = ".elf";

  public static final String TEXT_SECTION_NAME = ".text";
  public static final String DATA_SECTION_NAME = ".data";

  public static final String TEXT_SECTION_FILE_PREFIX = TEXT_SECTION_NAME + ".bin";
  public static final String DATA_SECTION_FILE_PREFIX = DATA_SECTION_NAME + ".bin";

  ElfProgramReader(final Model model, final Options options, final String fileName) {
    super(model, options, FileUtils.getShortFileNameNoExt(fileName),
        FileUtils.getShortFileNameNoExt(fileName) + TEXT_SECTION_FILE_PREFIX);

    this.fileName = fileName;

//    this.programFileNameNoExt = FileUtils.getShortFileNameNoExt(fileName);
//    this.programFileName = programFileNameNoExt + TEXT_SECTION_FILE_PREFIX;
  }

  public IrProgram getProgram() {
    try (final InputStream inputStream = new FileInputStream(fileName);
        final OutputStream outputInstructions = new FileOutputStream(programFileName);
        final OutputStream outputData =
            new FileOutputStream(fileName + DATA_SECTION_FILE_PREFIX);) {
      final ElfFile elfFile = ElfFile.from(inputStream);

      // .text
      final ElfSection instrSection = elfFile.firstSectionByName(TEXT_SECTION_NAME);
      Logger.debug("Reading %s section from ELF file. Address: 0x%s, size: 0x%s, offset: 0x%s.",
          TEXT_SECTION_NAME, Long.toHexString(instrSection.header.address),
          Long.toHexString(instrSection.header.size),
          Long.toHexString(instrSection.header.section_offset));

      ((FileInputStream) inputStream).getChannel().position(instrSection.header.section_offset);
      Logger.debug("Text section size: %s bytes.", (instrSection.header.size));
      final byte[] buffer = new byte[(int) instrSection.header.size];
      inputStream.read(buffer);
      outputInstructions.write(buffer);

      // Disassemble instructions
      final List<IsaPrimitive> program = disassembleProgram();

      // data
      final ElfSection dataSection = elfFile.firstSectionByName(DATA_SECTION_NAME);

      ((FileInputStream) inputStream).getChannel().position(dataSection.header.section_offset);
      final byte[] dataBuffer = new byte[(int) dataSection.header.size];
      inputStream.read(dataBuffer);
      outputData.write(dataBuffer);

      Logger.debug("Reading %s section from ELF file. Address: 0x%s, size: 0x%s, offset: 0x%s.",
          DATA_SECTION_NAME, Long.toHexString(dataSection.header.address),
          Long.toHexString(dataSection.header.size),
          Long.toHexString(dataSection.header.section_offset));

      return new IrProgram(program, instrSection.header.address, dataBuffer,
          dataSection.header.address, this.programFileNameNoExt);
    } catch (IOException ex) {
      ex.printStackTrace();
    }

    return null;
  }
}