/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.simulator;

import java.util.List;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.model.InstructionCall;
import ru.ispras.microtesk.model.IsaPrimitive;
import ru.ispras.microtesk.model.Model;
import ru.ispras.microtesk.model.ProcessingElement;
import ru.ispras.microtesk.model.memory.MemoryDevice;
import ru.ispras.microtesk.model.memory.loader.MemoryLoader;
import ru.ispras.microtesk.model.tracer.RecordFormatType;
import ru.ispras.microtesk.tools.simulator.memory.Element;
import ru.ispras.microtesk.tools.simulator.memory.IrMemory;
import ru.ispras.microtesk.tools.simulator.settings.SimulatorParameters;

/**
 * The {@code ProgramImage} class contains all information about the program to be simulated.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */

public class ProgramImage {
  public final static int DEFAULT_PE_NUMBER = 1;
  public final static int DEFAULT_ACTIVE_PE = 0;

  public final static int BIT_VECTOR_SIZE_64 = 64;

  private final Model model;
  private MemoryDevice pc;

  private ProcessingElement currentPe;

  private final String programName;
  private final List<IsaPrimitive> instructions;

  private final RecordFormatType tracerType;
  final SimulatorParameters simulatorParameters;

  public ProgramImage(final Model model, final IrProgram simulatorProgram,
      final IrMemory simulatorMemory, final SimulatorParameters simulatorParameters) {
    InvariantChecks.checkNotNull(model);
    InvariantChecks.checkNotNull(simulatorProgram);
    InvariantChecks.checkNotNull(simulatorParameters);

    Logger.debug("Instruction section address: %s.",
        Long.toHexString(simulatorProgram.getInstructionAddress()));

    initializeProcessingElement(model);
    this.model = model;

    // The current active ProcessingElement is used
    this.currentPe = model.getPE();

    ProgramImage.loadModelMemory(currentPe, simulatorMemory.getElements());

    this.simulatorParameters = simulatorParameters;
    this.programName = simulatorProgram.getProgramName();
    this.instructions  = simulatorProgram.getInstructions();

    // Store data
    this.storeProgramData(simulatorProgram, simulatorParameters.getMemName());

    this.pc = currentPe.getMemoryDevice(simulatorParameters.getPcName());
    this.pc.store(BitVector.valueOf(0, BIT_VECTOR_SIZE_64),
        BitVector.valueOf(simulatorParameters.getStartAddress(), BIT_VECTOR_SIZE_64));

    this.tracerType = RecordFormatType.getType(simulatorParameters.getTracerType());

    // TODO: remove
//    Set<String> storageSet = currentPe.getMemoryStorageSet();
//    storageSet.forEach(value -> Logger.debug("Storage name: %s", value));
  }

  private void initializeProcessingElement(final Model model) {
    initializeProcessingElement(model, DEFAULT_PE_NUMBER);
  }

  private void initializeProcessingElement(final Model model, final int number) {
    model.setPENumber(number);
    model.setActivePE(DEFAULT_ACTIVE_PE);
    Logger.debug("Processing element number: %s.", model.getPENumber());
  }

  private static void loadModelMemory(final ProcessingElement pe, final List<Element> elements) {
    for(Element element : elements) {
      MemoryLoader.loadMemoryElement(pe.getMemoryDevice(element.getMemoryName()), element.getAddress(), element.getValue());
    }
  }

  public long getPcValue() {
    return pc.load(BitVector.valueOf(0, BIT_VECTOR_SIZE_64)).longValue();
  }

  private void storeProgramData(final IrProgram simulatorProgram, final String dataMemName) {
    Logger.debug("Data section address: %s.",
        Long.toHexString(simulatorProgram.getDataAddress()));
    final byte[] data = simulatorProgram.getData();
    this.storeMemory64(dataMemName, simulatorProgram.getDataAddress(), data);
  }

  // Load 64 bits memory blocks
  // TODO: 16, 32, etc.
  private void storeMemory64(final String memoryName, long startDataAddress, final byte[] data) {
    // Load memory from current PE
    final MemoryDevice mem = currentPe.getMemoryDevice(memoryName);

    startDataAddress >>= 3;

    for (int j = 0; j < data.length / Long.BYTES; j++) {
      // ! bigEndian
      long result = 0;
      for (int i = 0; i < Long.BYTES; i++) {
        result <<= Byte.SIZE;
        result |= (data[i + j * Long.BYTES] & 0xFF);
      }

      mem.store(BitVector.valueOf(startDataAddress + j, BIT_VECTOR_SIZE_64), BitVector.valueOf(result, BIT_VECTOR_SIZE_64));

      Logger.debug("Memory (%s[0x%s]) addres: 0x%s <- 0x%s", memoryName,
          BitVector.valueOf((startDataAddress + j), BIT_VECTOR_SIZE_64).toHexString(),
          BitVector.valueOf((startDataAddress + j) << 3, BIT_VECTOR_SIZE_64).toHexString(),
          BitVector.valueOf(result, 64).toHexString());
    }
  }

  public List<IsaPrimitive> getInstructions() {
    return instructions;
  }

  public String getPathForFiles() {
    return simulatorParameters.getPath();
  }

  public String getProgramName() {
    return programName;
  }

  public int getExecutionLimit() {
    return simulatorParameters.getInstrLimit();
  }

  public String getFinalInstruction() {
    return simulatorParameters.getFinalInstr();
  }

  public int getMainStartAddress() {
    return simulatorParameters.getStartAddress();
  }

  public int getMainEndAddress() {
    return simulatorParameters.getEndAddress();
  }

  public String getFrameIndexRegister() {
    return simulatorParameters.getFrameIndexRegister();
  }

  public ProcessingElement getCurrentPe() {
    return currentPe;
  }

  public int getActivePE() {
    return model.getActivePE();
  }

  public InstructionCall getInstructionCall(final IsaPrimitive isaPrimitive) {
    return model.newCall(isaPrimitive);
  }

  public RecordFormatType getTracerType() {
    return tracerType;
  }
}
