/*
 * Copyright 2021-2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.simulator;

import java.io.IOException;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.model.InstructionCall;
import ru.ispras.microtesk.model.IsaPrimitive;
import ru.ispras.microtesk.model.ProcessingElement;
import ru.ispras.microtesk.model.memory.MemoryDevice;
import ru.ispras.microtesk.model.tracer.Record;
import ru.ispras.microtesk.model.tracer.RecordFactory;
import ru.ispras.microtesk.model.tracer.Tracer;
import ru.ispras.microtesk.test.template.ConcreteCall;
import ru.ispras.microtesk.tools.oracle.internal.Item;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class ProgramSimulator {
  // The field contains an object that describes all information about the program to be simulated.
  private final ProgramImage programImage;

  // PC
  private long currentPc;

  // The address of last instruction
  private long endPc = 0;

  // Active PE
  private ProcessingElement currentPe;

  private boolean terminateProgram = true;

  private Map<Long, InstructionCall> instructionMap;
  private Map<Long, ConcreteCall> concreteInstructionMap;

  public ProgramSimulator(final ProgramImage programImage) {
    InvariantChecks.checkNotNull(programImage);

    this.programImage = programImage;

    currentPe = programImage.getCurrentPe();

    // Prepare instruction
    loadInstruction();

    // Prepare PC
    // TODO
    currentPc = programImage.getPcValue();
    Logger.debug("[%s] PC start address: 0x%s", ProgramSimulator.class.getSimpleName(),
        Long.toHexString(currentPc));

    terminateProgram = false;
  }

  private void loadInstruction() {
    instructionMap = new HashMap<Long, InstructionCall>();
    concreteInstructionMap = new HashMap<Long, ConcreteCall>(); // TODO:
    long mapPc = 0;
    List<IsaPrimitive> instructions = programImage.getInstructions();
    for (final IsaPrimitive isaPrimitive : instructions) {
      InstructionCall call = programImage.getInstructionCall(isaPrimitive);

      Logger.debug("Load instruction: [PC] 0x%s <- %s (0x%s)", Long.toHexString(mapPc), call.getText(),
          (new BigInteger(call.getImage(), 2)).toString(16));

      instructionMap.put(mapPc, call);

      ConcreteCall tempConcreteCall = new ConcreteCall(call);
      tempConcreteCall.setAddress(BigInteger.valueOf(mapPc)); // Adds address to instruction for tracer
      concreteInstructionMap.put(mapPc, tempConcreteCall); // TODO:

      mapPc += call.getByteSize();
    }

    endPc = mapPc;
  }

  public boolean hasNext() {
    return (currentPc != endPc) && !terminateProgram;
  }

  public ConcreteCall nextInstruction() {
    return concreteInstructionMap.get(currentPc);
  }

  public String nextInstructionText() {
    return concreteInstructionMap.get(currentPc).getText();
  }

  /**
   * Executes the {@code ConcreteCall} and if it was the last program instruction changes the state
   * {@code terminateProgram} to {@code true}.
   */
  public void nextStep() {
    ConcreteCall concreteCall = concreteInstructionMap.get(currentPc);

    // Next step: Pre
    if (concreteCall == null) {
      Logger.debug("The end of the program has been reached: 0x%s.", Long.toHexString(currentPc));
      terminateProgram = true;
      return;
    } else if (concreteCall.getText().equals(programImage.getFinalInstruction())
        && checkMainAddressRange()) {
      terminateProgram = true;
    } else if (concreteCall.getExecutionCount() > programImage.getExecutionLimit()) {
      terminateProgram = true;
      throw new RuntimeException(String.format(
          "Instruction %s reached its limit on execution count (%d). "
              + "Probably, the program entered an endless loop. Simulation was aborted.",
          concreteCall.getText(), programImage.getExecutionLimit()));
    }

    // Next step: Action
    // Trace log
    final Record record = RecordFactory.newInstruction(concreteCall, programImage.getActivePE());
    Tracer.addRecord(record);

    Logger.debug("[%s] PC: 0x%s, call: %s", ProgramSimulator.class.getSimpleName(),
        Long.toHexString(currentPc), concreteCall.getText());
    concreteCall.execute(currentPe);
    // debug
    // printRegister("GPR", 8, pe);
    // printRegisters("GPR", pe);

    currentPc = programImage.getPcValue();
  }

  // Checks if the current instruction PC address is in the main program address range.
  private boolean checkMainAddressRange() {
    final int minAddress = programImage.getMainStartAddress();
    final int maxAddress = programImage.getMainEndAddress();
    if (maxAddress - minAddress != 0) {
      if (maxAddress >= currentPc && currentPc >= minAddress) {
        return true;
      }
      return false;
    }
    return true;
  }

  public long getRegisterValue(final Item registerMode, final int number) {
    final String registerName = registerMode.getItemName();
    long frameShift = 0;
   // Logger.debug("======= programImage.getFrameIndexRegister() [%s]", programImage.getFrameIndexRegister());
    if(!programImage.getFrameIndexRegister().isEmpty() && registerMode.getFrameDependence()) {
      final MemoryDevice frameIndexRegister = currentPe.getMemoryDevice(programImage.getFrameIndexRegister());
      frameShift = frameIndexRegister.load(BitVector.valueOf(0, ProgramImage.BIT_VECTOR_SIZE_64)).longValue();
      //Logger.debug("======= frameShift [%s]", frameShift);
    }

    final MemoryDevice registers = currentPe.getMemoryDevice(registerName);
    final int size = registers.getAddressBitSize();
    final int registerIndex = (int) (number + frameShift*(registerMode.getEndIndex() + 1));

//    Logger.debug("======= registerIndex [%s]", registerIndex);
//    Logger.debug("======= number: [%s]", number);
//    Logger.debug("======= registerName: [%s]", registerName);

    if (registerIndex < (1 << size)) {
      return registers.load(BitVector.valueOf(registerIndex, 64)).longValue();
    }
    return 0;
  }

  public boolean simulate() {
    Tracer.initialize(programImage.getPathForFiles(), programImage.getProgramName());

    // The trace log is always created
    Tracer.setEnabled(true);

    // Set the style of printed events to trace
    RecordFactory.setRecordType(programImage.getTracerType());

    try {
      Tracer.createFile();

      // debug
//      if (simulatorProgram.getData().length > 0) {
//        MemoryDevice mem = pe.getMemoryDevice(memoryName);
//        int size = simulatorProgram.getData().length / 8; // TODO:
//        printMemory(mem, simulatorProgram.getDataAddress(), size);
//      }
      while (hasNext()) {
        nextStep();
      }
      Logger.debug("[ProgramSimulator] Simulation finished.");

      Tracer.closeFile(); // TODO
    } catch (IOException e) {
      e.printStackTrace();
    }

    return true;
  }

  public long getCurrentPc() {
    return currentPc;
  }
}
