/*
 * Copyright 2021-2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.simulator;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.model.Model;
import ru.ispras.microtesk.options.Options;
import ru.ispras.microtesk.tools.ToolUtils;
import ru.ispras.microtesk.tools.simulator.memory.IrMemory;
import ru.ispras.microtesk.tools.simulator.memory.MemoryFactory;
import ru.ispras.microtesk.tools.simulator.settings.SimulatorParameters;

/**
 * The {@code Simulator} class simulates the input program on the target model.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public final class Simulator {

  /**
   * Simulates a program on the microprocessor model.
   *
   * @param options the simulator settings.
   * @param modelName the microprocessor model name.
   * @param inputProgram the binary file name.
   * @param inputResourceDir the path to resource directory.
   * @return result of simulation: {@code true} if success and {@code false} if failed.
   */
  public static boolean simulate(final Options options, final String modelName,
      final String inputProgram, final String inputResourceDir) {
    InvariantChecks.checkNotNull(options);
    InvariantChecks.checkNotNull(modelName);
    InvariantChecks.checkNotNull(inputProgram);

    // TODO: Debug flag
    Logger.setDebug(true);

    final Model model = ToolUtils.loadModel(modelName);

    // Files for simulator
    final SimulatorSettings simulatorSettings =
        new SimulatorSettings(inputResourceDir, inputProgram, options);

    // Disassemble
    final ProgramReader programReader = ProgramReaderFactory.getBinaryProgramReader(model, options,
        simulatorSettings.getProgramFileName());
    final IrProgram simulatorProgram = programReader.getProgram();

    // Options for simulator
    final SimulatorParameters simulatorParameters = simulatorSettings.getSimulatorParameters();
    // Logger.debug
    SimulatorParameters.printParameters(Simulator.class.getSimpleName(), simulatorParameters);

    // Get memory image
    final IrMemory simulatorMemory =
        MemoryFactory.newMemory(modelName, simulatorSettings.getMemoryFiles());

    // Simulate
    ProgramImage programImage =
        new ProgramImage(model, simulatorProgram, simulatorMemory, simulatorParameters);
    final ProgramSimulator programSimulator = new ProgramSimulator(programImage);

    return programSimulator.simulate();
  }
}
