/*
 * Copyright 2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.simulator;

import java.io.File;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

import javax.json.JsonObject;
import javax.json.JsonReader;

import ru.ispras.castle.util.Logger;
import ru.ispras.microtesk.options.Options;
import ru.ispras.microtesk.tools.simulator.memory.MemoryFiles;
import ru.ispras.microtesk.tools.simulator.settings.SimulatorParameters;
import ru.ispras.microtesk.tools.simulator.settings.SimulatorParametersFactory;
import ru.ispras.microtesk.utils.JsonConfig;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class SimulatorSettings {
  // The name of the file with the simulator settings
  public static final String SIMULATOR_SETTINGS = "simulator_settings.json";

  // Default settings for all programs of target architecture
  public static final String SIMULATOR_SETTINGS_DEFAULT = "simulator_program_settings_default.json";

  // Setting for simulator
  public static final String FILE_POSTFIX_PROGRAM_OPTIONS = "file_postfix_program_options";

  // Setting for program
  public static final String FILE_MAIN_PROGRAM_ADDRESS = "file_main_address";
  public static final String FILE_MEMORY_PATHS = "file_memory_paths";

  private final String inputResourceDir;

  private final String programFileName;
  private final Path programFilePath;
  private final Path programFileDirPath;

  private final String programFileOptions;

  private final String programFileMainAddress;
  private final String programFileMemoryPaths;

  private final Options options;

  public SimulatorSettings(final String inputResourceDir, final String programFileName, final Options options) {
    this.inputResourceDir = inputResourceDir;

    this.programFileName = programFileName;
    this.programFilePath = Paths.get(programFileName);
    if (!Files.exists(programFilePath)) {
      Logger.error("[SimulatorSettings] Files %s is not exist.", programFileName);
      //throw new FileNotFoundException(); 
    }
    this.programFileDirPath = programFilePath.getParent();

    this.programFileOptions = programFileName + this.getPostfixProgramOptionsFileName(inputResourceDir);
    //Logger.debug("this.programFileOptions: %s", this.programFileOptions);

    {
      final Path programFileOptionsPath = Paths.get(programFileOptions);
      if (!Files.exists(programFileOptionsPath)) {
        Logger.error("[SimulatorSettings] Files %s is not exist.", programFileOptionsPath);
        //throw new FileNotFoundException(); 
      }

      JsonReader jsonReader = JsonConfig.getJsonReader(programFileOptionsPath.toString());
      JsonObject jsonObject = jsonReader.readObject();

      this.programFileMainAddress = jsonObject.getString(FILE_MAIN_PROGRAM_ADDRESS);
      this.programFileMemoryPaths = jsonObject.getString(FILE_MEMORY_PATHS);
    }
    Logger.debug("[SimulatorSettings] File for the program that contains main address: %s", this.programFileMainAddress);
    Logger.debug("[SimulatorSettings] File for the program that contains paths to the memory files: %s", this.programFileMemoryPaths);

    this.options = options;
  }

  private String getPostfixProgramOptionsFileName(final String inputResourceDir) {
    final String simulatorSettings = inputResourceDir + File.separator + SIMULATOR_SETTINGS;

    final Path simulatorSettingsPath = Paths.get(simulatorSettings);
    if (!Files.exists(simulatorSettingsPath)) {
      Logger.error("[SimulatorSettings] Files %s is not exist.", simulatorSettingsPath);
      // throw new FileNotFoundException(); 
    }

    JsonReader jsonReader = JsonConfig.getJsonReader(simulatorSettingsPath.toString());
    JsonObject jsonObject = jsonReader.readObject();
    final String postfixProgramOptions = jsonObject.getString(FILE_POSTFIX_PROGRAM_OPTIONS);
    JsonConfig.close(jsonReader);

    return postfixProgramOptions;
  }

  public String getProgramFileName() {
    return programFileName;
  }

  public MemoryFiles getMemoryFiles() {
    return new MemoryFiles(this.getMemoryPathsFile(), this.getSourceDirectory());
  }

  public Path getMemoryPathsFile() {
    return programFileDirPath.resolve(programFileMemoryPaths);
  }

  public Path getSourceDirectory() {
    return programFileDirPath;
  }

  public SimulatorParameters getSimulatorParameters() {
    final Path pathOptions = Paths.get(this.inputResourceDir).resolve(SIMULATOR_SETTINGS_DEFAULT);

    final SimulatorParameters simulatorParameters =
        SimulatorParametersFactory.newSimulatorParameters(pathOptions,
            programFileDirPath.resolve(programFileMainAddress), options);

    Logger.debug("pathOptions: %s", pathOptions);
    Logger.debug("programFileDirPath.resolve(programFileMainAddress): %s", programFileDirPath.resolve(programFileMainAddress));

    return updateParameters(simulatorParameters);
  }

  private SimulatorParameters updateParameters(final SimulatorParameters simulatorParameters) {
    JsonReader jsonReader = JsonConfig.getJsonReader(programFileOptions);
    JsonObject jsonObject = jsonReader.readObject();

    if (jsonObject.containsKey(SimulatorParametersFactory.PARAMETER_EXECUTION_LIMIT)) {
      final int instructionExecutionLimit =
          jsonObject.getInt(SimulatorParametersFactory.PARAMETER_EXECUTION_LIMIT);
      Logger.debug("instructionExecutionLimit: %s", instructionExecutionLimit);
      final SimulatorParameters simulatorParametersUpd =
          new SimulatorParameters(simulatorParameters, instructionExecutionLimit);
      JsonConfig.close(jsonReader);

      return simulatorParametersUpd;
    } else {
      JsonConfig.close(jsonReader);

      return simulatorParameters;
    }
  }
}