/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.simulator.memory;

import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Map;
import java.util.Set;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.data.types.bitvector.BitVector;
import ru.ispras.microtesk.model.memory.loader.MemoryRegion;
import ru.ispras.microtesk.model.memory.loader.MemoryStructure;
import ru.ispras.microtesk.model.memory.loader.ParametersLoader;
import ru.ispras.microtesk.utils.BitStream;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class ArkIrMemory extends IrMemory {

  /**
   * Creates a memory image for the simulator
   *
   * @param modelName the model name.
   * @param memoryFiles the paths for memory files.
   */
  public ArkIrMemory(final String modelName, final MemoryFiles memoryFiles) {

    // List of files to be loaded into memory
    Map<String, String> files = IrMemory.readMemoryFiles(memoryFiles.getMemoryPathsFile());
    files.forEach((key, value) -> Logger.debug("key: %s, value: %s", key,  value));

    // Memory record structure read from file
    Path memoryStructureFile = ParametersLoader.getStructurePath(modelName);
    Map<String, MemoryStructure> structure = MemoryStructure.loadMemoryElementStructure(memoryStructureFile.toString());
    structure.forEach((key, value) -> Logger.debug("key: %s, value: [%s]", key,  value));

    // Memory regions read from file
    Path memoryRegionsFile = ParametersLoader.getRegionsPath(modelName);
    Set<MemoryRegion> regions = MemoryRegion.loadMemoryRegions(memoryRegionsFile.toString());
    regions.forEach(value -> Logger.debug("regions value: [%s]", value));

    //List<Element> elements = new ArrayList<>();
    // TODO:
    for(MemoryRegion region : regions) {
      MemoryStructure targetStructure = structure.get(region.getType());

      int address = region.getStartIndex();
      final String memName = region.getMemoryName();
      final String memType = region.getType();

      final String memFileName = files.get(region.getType());
      if(null == memFileName) {
        continue;
      }
      final Path pathMemFileName = memoryFiles.getSourceDirectory().resolve(memFileName);
      if(Files.exists(pathMemFileName)) {
        Logger.debug("Loads nemory elements from: %s", pathMemFileName);
        try (final InputStream source = Files.newInputStream(pathMemFileName)) {
          BitStream bitStream = BitStream.of(source);

          // TODO:
          while(!bitStream.isEmpty()) {
            for(String elementName: targetStructure.getElements()) {
              BitVector value = bitStream.read(targetStructure.getDataSize());

              elements.add(new Element(address, value, memName, memType, elementName));

              address++;
            }
          }
        } catch (IOException e) {
          e.printStackTrace();
        }
      }
      else {
        Logger.error("The specified memory file is missing: %s", pathMemFileName.toString());
      }
    }
    elements.forEach(value -> Logger.debug("elements: [%s]", value));
  }
}
