/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.simulator.settings;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.model.memory.loader.MemoryImageLoader;
import ru.ispras.microtesk.options.Option;
import ru.ispras.microtesk.options.Options;
import ru.ispras.microtesk.tools.ToolUtils;

import java.io.IOException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;

import static ru.ispras.microtesk.model.memory.loader.MemoryImageLoader.Request;

public final class SimulatorOptions extends SimulatorParameters {
  private final List<Request> preloadRequests;

  public SimulatorOptions(final Options options) {
    super(options.getValueAsString(Option.PROGRAM_COUNTER_NAME), // programCounterName
          options.getValueAsString(Option.MEMORY_NAME), // memoryName
          ToolUtils.getOutDir(options), // pathForFiles
          options.getValueAsInteger(Option.INSTRUCTION_SIMULATION_LIMIT), // instructionExecutionLimit
          ToolUtils.getFinalInstruction(options), // finalInstruction
          options.getValueAsString(Option.TRACER_RECORD_TYPE), // tracerType
          0, 0, "");

    preloadRequests = parseMemoryImagePreloadOptions(options);
  }

  public static void printOptions(final String toolName, final SimulatorOptions options) {
    InvariantChecks.checkNotNull(options);

    Logger.debug(String.format("%s settings: %s", toolName, options.toString()));
  }

  public List<Request> getPreloadRequests() {
    return preloadRequests;
  }

  private static List<Request> parseMemoryImagePreloadOptions(Options opts) {
    final var commandLine = opts.getCommandLine();

    final var source = new ImageDetails();
    final var target = new ImageDetails();
    final var reqs = new ArrayList<Request>();

    for (var opt: commandLine.getOptions()) {
      final String name = opt.getLongOpt();
      if (Option.PRELOAD_TARGET.getName().equals(name)) {
        if (target.readOption(opt) && source.isSet()) {
          makeRequest(reqs, target, source);
        }
      }
      if (Option.PRELOAD_SOURCE.getName().equals(name)) {
        if (source.readOption(opt) && target.isSet()) {
          makeRequest(reqs, target, source);
        }
      }
    }
    return reqs;
  }

  private static void makeRequest(
      List<Request> reqs, ImageDetails target, ImageDetails source) {
    try {
      if (target.hi > 0) {
        reqs.add(MemoryImageLoader.request(
            target.name, target.lo, target.hi, Path.of(source.name), source.lo));
      } else {
        reqs.add(MemoryImageLoader.request(
            target.name, target.lo, Path.of(source.name), source.lo, source.hi));
      }
      target.reset();
      source.reset();
    } catch (IOException e) {
      Logger.warning(PRELOAD_SOURCE_WARNING, source.name, e.getMessage());
    }
  }

  private static class ImageDetails {
    String name;
    long lo;
    long hi;

    ImageDetails() {
      reset();
    }

    void reset() {
      name = "";
      lo = -1;
      hi = -1;
    }

    boolean isSet() {
      return !name.isEmpty();
    }

    boolean readOption(org.apache.commons.cli.Option opt) {
      final String optName = opt.getLongOpt();
      final String[] values = opt.getValues();
      if (values == null) {
        reset();
      } else try {
        long offsetLo = 0, offsetHi = -1;
        if (values.length > 1) {
          offsetLo = parseLong(values[1]);
          if (offsetLo < 0) {
            Logger.warning(PRELOAD_LO_WARNING, values[1], optName);
            return false;
          }
        }
        if (values.length > 2) {
          offsetHi = parseLong(values[2]);
          if (offsetHi <= offsetLo) {
            Logger.warning(PRELOAD_HI_WARNING, values[2], optName);
            return false;
          }
        }

        this.name = values[0];
        this.lo = offsetLo;
        this.hi = offsetHi;

        return true;
      } catch (NumberFormatException e) {
        Logger.warning(PRELOAD_FORMAT_WARNING, optName);
      }
      return false;
    }
  }

  private static long parseLong(String s) {
    if (s.startsWith("0x")) {
      return Long.parseLong(s, 2, s.length(), 16);
    }
    return Long.parseLong(s);
  }

  private static String PRELOAD_SOURCE_WARNING =
    "Ignoring the option to load an image from %s, reason: %s";

  private static String PRELOAD_LO_WARNING =
    "Low offset argument %s for the option %s is negative, ignoring the option";

  private static String PRELOAD_HI_WARNING =
    "High offset argument %s for the option %s is not greater than low offset, ignoring the option";

  private static String PRELOAD_FORMAT_WARNING =
    "An offset argument for the option %s has wrong number format, ignoring the option";
}
