/*
 * Copyright 2022-2023 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.simulator.settings;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.options.Option;
import ru.ispras.microtesk.options.Options;
import ru.ispras.microtesk.tools.ToolUtils;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class SimulatorParameters {
  protected final String programCounterName;
  protected final String memoryName;
  protected final String pathOutDir;
  protected final int instructionExecutionLimit;
  protected final String finalInstruction;
  protected final String tracerType;
  protected final int startAddress;
  protected final int endAddress;
  protected final String frameIndexRegister;

  public static SimulatorParameters newSimulatorParameters(final Options options) {
    InvariantChecks.checkNotNull(options);

    return new SimulatorParameters(options.getValueAsString(Option.PROGRAM_COUNTER_NAME), // programCounterName
        options.getValueAsString(Option.MEMORY_NAME), // memoryName
        ToolUtils.getOutDir(options), // pathForFiles
        options.getValueAsInteger(Option.INSTRUCTION_SIMULATION_LIMIT), // instructionExecutionLimit
        ToolUtils.getFinalInstruction(options), // finalInstruction
        options.getValueAsString(Option.TRACER_RECORD_TYPE),
        0, // startAddress
        0, // endAddress
        "");
  }

  protected SimulatorParameters(final String pc, final String mem, final String path,
      final int instrLimit, final String finalIntr, final String tracer, final int startAddress,
      final int endAddress, final String frameIndexRegister) {
    this.programCounterName = pc;
    this.memoryName = mem;
    this.pathOutDir = path;
    this.instructionExecutionLimit = instrLimit;
    this.finalInstruction = finalIntr;
    this.tracerType = tracer;
    this.startAddress = startAddress;
    this.endAddress = endAddress;
    this.frameIndexRegister = frameIndexRegister;
  }

  public SimulatorParameters(final SimulatorParameters simulatorParameters, final int instrLimit) {
    this.programCounterName = simulatorParameters.programCounterName;
    this.memoryName = simulatorParameters.memoryName;
    this.pathOutDir = simulatorParameters.pathOutDir;
    this.instructionExecutionLimit = instrLimit;
    this.finalInstruction = simulatorParameters.finalInstruction;
    this.tracerType = simulatorParameters.tracerType;
    this.startAddress = simulatorParameters.startAddress;
    this.endAddress = simulatorParameters.endAddress;
    this.frameIndexRegister = simulatorParameters.frameIndexRegister;
  }

  protected SimulatorParameters(final String pc, final String mem, final String path,
      final int instrLimit, final String finalIntr, final String tracer,
      final String frameIndexRegister) {
    this(pc, mem, path, instrLimit, finalIntr, tracer, 0, 0, frameIndexRegister);
  }

  /**
   * @return the program counter name
   */
  public String getPcName() {
    return programCounterName;
  }

  /**
   * @return the memory name
   */

  public String getMemName() {
    return memoryName;
  }

  /**
   * @return the path for output simulator files
   */
  public String getPath() {
    return pathOutDir;
  }

  /**
   * @return the instruction execution limit
   */
  public int getInstrLimit() {
    return instructionExecutionLimit;
  }

  /**
   * @return the final instruction
   */
  public String getFinalInstr() {
    return finalInstruction;
  }

  /**
   * @return the tracer printer type
   */
  public String getTracerType() {
    return tracerType;
  }

  /**
   * @return the main program start address
   */
  public int getStartAddress() {
    return startAddress;
  }

  /**
   * @return the main program end address
   */
  public int getEndAddress() {
    return endAddress;
  }

  /**
   * @return the name of the register containing the frame index
   */
  public String getFrameIndexRegister() {
    return frameIndexRegister;
  }

  @Override
  public String toString() {
    return String.format(
        "PC:[%s], Memory:[%s], Path:[%s], One instruction simulation limit:[%s], Final instruction:[%s], trace type:[%s], frame index reg:[%s]",
        programCounterName, memoryName, pathOutDir, instructionExecutionLimit, finalInstruction,
        tracerType, frameIndexRegister);
  }

  public static void printParameters(final String toolName, final SimulatorParameters parameters) {
    InvariantChecks.checkNotNull(parameters);

    Logger.debug(String.format("%s settings: %s", toolName, parameters.toString()));
  }
}
