/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.simulator.settings;

import java.nio.file.Files;
import java.nio.file.Path;

import javax.json.JsonObject;
import javax.json.JsonReader;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.options.Options;
import ru.ispras.microtesk.tools.ToolUtils;
import ru.ispras.microtesk.utils.JsonConfig;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class SimulatorParametersFactory {
  public static final String PARAMETER_PC_NAME = "pc_name";
  public static final String PARAMETER_MEMORY_NAME = "memory_name";
  public static final String PARAMETER_OUT_DIR = "path_out_dir";
  public static final String PARAMETER_EXECUTION_LIMIT = "instruction_execution_limit";
  public static final String PARAMETER_FINAL_INSTRUCTION = "final_instruction";
  public static final String PARAMETER_TRACE_TYPE = "tracer_type";
  public static final String PARAMETER_START_ADDRESS = "start_address";
  public static final String PARAMETER_END_ADDRESS = "end_address";
  // Number of registers in a frame
  public static final String PARAMETER_FRAME_INDEX_REGISTER = "frame_index_register";

  public static SimulatorParameters newSimulatorParameters(final Path optionsFilePath, final Options options) {
    InvariantChecks.checkNotNull(optionsFilePath);
    InvariantChecks.checkNotNull(options);

    if (Files.exists(optionsFilePath)) {
      Logger.debug(String.format("[Simulator] Files %s is exist.", optionsFilePath));
      return newSimulatorParametersJson(optionsFilePath);
    } else {
      Logger.debug(String.format("[Simulator] Files %s is not exist.", optionsFilePath));
      return SimulatorParameters.newSimulatorParameters(options);
    }
  }

  public static SimulatorParameters newSimulatorParameters(final Path optionsFilePath, final Path addressMainProgram, final Options options) {
    InvariantChecks.checkNotNull(optionsFilePath);
    InvariantChecks.checkNotNull(options);

    if (Files.exists(optionsFilePath)) {
      Logger.debug(String.format("[Simulator] Files %s is exist.", optionsFilePath));

      if (Files.exists(addressMainProgram)) {
        Logger.debug(String.format("[Simulator] Files %s is exist.", addressMainProgram));
        return newSimulatorParametersJson(optionsFilePath, addressMainProgram);
      } else {
        Logger.debug(String.format("[Simulator] Files %s is not exist.", addressMainProgram));
        return newSimulatorParametersJson(optionsFilePath);
      }
    } else {
      Logger.debug(String.format("[Simulator] Files %s is not exist.", optionsFilePath));
      return SimulatorParameters.newSimulatorParameters(options);
    }
  }

  /*
    The example json file with parameters:
    {
      "pc_name": "PC",
      "memory_name": "MEM",
      "path_out_dir": "",
      "instruction_execution_limit": 100,
      "final_instruction": "return",
      "tracer_type": "ark",
      "start_address": "13",
      "end_address": "3e",
      "frame_index_register":"FC"
    }
   */
  public static SimulatorParameters newSimulatorParametersJson(final Path optionsFilePath,
      final Path addressMainProgram) {
    InvariantChecks.checkNotNull(addressMainProgram);

    JsonReader jsonReaderAddress = JsonConfig.getJsonReader(addressMainProgram.toString());
    JsonObject jsonAddressObject = jsonReaderAddress.readObject();
//    Integer startAddress = Integer.parseInt(jsonAddressObject.getString(PARAMETER_START_ADDRESS), 16);
//    Integer endAddress = Integer.parseInt(jsonAddressObject.getString(PARAMETER_END_ADDRESS), 16);
    Integer startAddress = Integer.decode(jsonAddressObject.getString(PARAMETER_START_ADDRESS));
    Integer endAddress = Integer.decode(jsonAddressObject.getString(PARAMETER_END_ADDRESS));
    JsonConfig.close(jsonReaderAddress);

    InvariantChecks.checkNotNull(optionsFilePath);
    JsonReader jsonReader = JsonConfig.getJsonReader(optionsFilePath.toString());
    JsonObject jsonObject = jsonReader.readObject();

    SimulatorParameters simulatorParameters =
        new SimulatorParameters(jsonObject.getString(PARAMETER_PC_NAME), // programCounterName
            jsonObject.getString(PARAMETER_MEMORY_NAME), // memoryName
            ToolUtils.getOutDir(jsonObject.getString(PARAMETER_OUT_DIR)), // pathForFiles
            jsonObject.getInt(PARAMETER_EXECUTION_LIMIT), // instructionExecutionLimit
            jsonObject.getString(PARAMETER_FINAL_INSTRUCTION), // finalInstruction
            jsonObject.getString(PARAMETER_TRACE_TYPE), startAddress, // startAddress
            endAddress, // endAddress
            jsonObject.getString(PARAMETER_FRAME_INDEX_REGISTER));

    JsonConfig.close(jsonReader);

    return simulatorParameters;
  }

  public static SimulatorParameters newSimulatorParametersJson(final Path optionsFilePath) {
    InvariantChecks.checkNotNull(optionsFilePath);

    JsonReader jsonReader = JsonConfig.getJsonReader(optionsFilePath.toString());

    JsonObject jsonObject = jsonReader.readObject();

    SimulatorParameters simulatorParameters =
        new SimulatorParameters(jsonObject.getString(PARAMETER_PC_NAME), // programCounterName
            jsonObject.getString(PARAMETER_MEMORY_NAME), // memoryName
            ToolUtils.getOutDir(jsonObject.getString(PARAMETER_OUT_DIR)), // pathForFiles
            jsonObject.getInt(PARAMETER_EXECUTION_LIMIT), // instructionExecutionLimit
            jsonObject.getString(PARAMETER_FINAL_INSTRUCTION), // finalInstruction
            jsonObject.getString(PARAMETER_TRACE_TYPE),
            jsonObject.getString(PARAMETER_FRAME_INDEX_REGISTER));

    JsonConfig.close(jsonReader);

    return simulatorParameters;
  }
}
