/*
 * Copyright 2017-2021 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.templgen;

import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Set;

import ru.ispras.castle.util.Logger;
import ru.ispras.fortress.util.InvariantChecks;
import ru.ispras.microtesk.SysUtils;
import ru.ispras.microtesk.model.Model;
import ru.ispras.microtesk.model.metadata.MetaModel;
import ru.ispras.microtesk.options.Option;
import ru.ispras.microtesk.options.Options;
import ru.ispras.microtesk.tools.ToolUtils;
import ru.ispras.microtesk.tools.templgen.printers.RubyTemplatePrinter;
import ru.ispras.microtesk.tools.templgen.templates.AllPathsTemplate;
import ru.ispras.microtesk.tools.templgen.templates.BoundaryValuesTemplate;
import ru.ispras.microtesk.tools.templgen.templates.GroupTemplate;
import ru.ispras.microtesk.tools.templgen.templates.SequenceTemplate;
import ru.ispras.microtesk.tools.templgen.templates.SimpleTemplate;

/**
 * The {@code TemplateGenerator} class automatically creates a set of templates for a model of
 * microprocessor.
 *
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 * @author <a href="mailto:andrewt@ispras.ru">Andrei Tatarnikov</a>
 */
public final class TemplateGenerator {

  /**
   * Generates a set of templates for a model of microprocessor.
   *
   * @param options the generation settings.
   * @param modelName the microprocessor model name.
   * @return result of generation: {@code true} if success and {@code false} if failed.
   */
  public static boolean generate(final Options options, final String modelName) {
    InvariantChecks.checkNotNull(options);
    InvariantChecks.checkNotNull(modelName);

    Logger.message("Starting the template generation tool.");

    String baseTemplateName = options.getValueAsString(Option.BASE_TEMPLATE_NAME);
    if (baseTemplateName.isEmpty()) {
      Logger.warning("Failed to get the base template name for the '%s' model."
          + " The base template was named as model.", modelName);
      baseTemplateName = modelName;
    }

    String baseTemplatePath = options.getValueAsString(Option.BASE_TEMPLATE_PATH);
    if (baseTemplatePath.isEmpty()) {
      Logger.warning("Failed to get the base template path for the '%s' model."
          + " The base template path was named by default.", modelName);
      baseTemplatePath = modelName;
    }

    final Path outputDirectory;
    if (options.hasValue(Option.OUTPUT_DIR)) {
      outputDirectory = Paths.get(options.getValueAsString(Option.OUTPUT_DIR));
    } else if (options.hasValue(Option.ARCH_DIRS)) {
      final String archDirs = options.getValueAsString(Option.ARCH_DIRS);
      final Path archPath = SysUtils.searchArchSettingsPath(archDirs, modelName);
      outputDirectory = archPath.getParent().resolve("templates");
    } else {
      Logger.error("The --%s option is undefined.", Option.ARCH_DIRS.getName());
      outputDirectory = Paths.get("");
    }

    Logger.message("Output directory: %s", outputDirectory);

    final Model model = ToolUtils.loadModel(modelName);
    final MetaModel metaModel = model.getMetaData();
    final Set<String> ignoredInstructions = getIgnoredInstructions(options);

    boolean generatedResult = true;

    // TODO: Scenario iterator
    if (options.hasValue(Option.TEMPLATE_SCENARIO_NAME)) {
      final Set<String> scenarios = options.getValueAsStringSet(Option.TEMPLATE_SCENARIO_NAME);
      if (scenarios.contains(AllPathsTemplate.ALL_PATHS_TEMPLATE_TARGET)) {
        final AllPathsTemplate template =
            new AllPathsTemplate(
                metaModel, new RubyTemplatePrinter(AllPathsTemplate.ALL_PATHS_TEMPLATE_NAME,
                    modelName, baseTemplateName, baseTemplatePath, outputDirectory, false),
                ignoredInstructions);
        generatedResult &= template.generate();
      }
    } else {
      final SimpleTemplate simpleTemplate =
          new SimpleTemplate(metaModel, new RubyTemplatePrinter(SimpleTemplate.SIMPLE_TEMPLATE_NAME,
              modelName, baseTemplateName, baseTemplatePath, outputDirectory, true), ignoredInstructions);
      generatedResult &= simpleTemplate.generate();

      final GroupTemplate groupTemplate =
          new GroupTemplate(metaModel, new RubyTemplatePrinter(GroupTemplate.GROUP_TEMPLATE_NAME,
              modelName, baseTemplateName, baseTemplatePath, outputDirectory, true), ignoredInstructions);
      generatedResult &= groupTemplate.generate();

      final BoundaryValuesTemplate boundaryTemplate =
          new BoundaryValuesTemplate(
              metaModel, new RubyTemplatePrinter(BoundaryValuesTemplate.BOUNDARY_TEMPLATE_NAME,
                  modelName, baseTemplateName, baseTemplatePath, outputDirectory, true),
              ignoredInstructions);
      generatedResult &= boundaryTemplate.generate();

      // TODO: Fix: copy of boundary
      final SequenceTemplate sequenceTemplate = new SequenceTemplate(metaModel,
          new RubyTemplatePrinter(SequenceTemplate.SEQUENCE_TEMPLATE_NAME, modelName,
              baseTemplateName, baseTemplatePath, outputDirectory, true),
          ignoredInstructions);
      generatedResult &= sequenceTemplate.generate();
    }

    Logger.message("Template generation complete. Result: %s", generatedResult ? "Successed" : "Failure"); // TODO:

    return generatedResult;
  }

  /**
   * Returns a set of instructions.
   *
   * @param options the generation settings.
   * @return the set of instructions to ignore.
   */
  private static Set<String> getIgnoredInstructions(final Options options) {
    return options.getValueAsStringSet(Option.IGNORED_INSTRUCTIONS);
  }
}
