package ru.ispras.microtesk.translator.mir;

import ru.ispras.castle.util.Logger;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardCopyOption;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

public final class Callgraph {
  private final CallVisitor visitor = new CallVisitor();
  private final Map<String, Set<String>> calledBy = new HashMap<>();
  private final Map<String, MirContext> callers = new HashMap<>();

  public Collection<MirContext> getRecursive() {
    return calledBy.entrySet().stream()
      .filter(e -> e.getValue().contains(e.getKey()))
      .map(Map.Entry::getKey)
      .map(callers::get)
      .collect(Collectors.toList());
  }

  public void addAll(Collection<MirContext> mirs) {
    for (var mir: mirs) {
      add(mir);
    }
  }

  public void add(MirContext ctx) {
    visitor.callees =
        calledBy.computeIfAbsent(ctx.name, k -> new HashSet<>());
    visitor.visitBlocks(ctx.blocks);

    if (!visitor.callees.isEmpty()) {
      callers.put(ctx.name, ctx);
    }
  }

  public void write(Path path) {
    Path temp = null;
    try {
      temp = Files.createTempFile("callgraph", ".dot");
      try (final var writer = new DotWriter(temp)) {
        writer.start("Callgraph");
        for (var entry: calledBy.entrySet()) {
          writer.writeAll(entry.getKey(), entry.getValue());
        }
        writer.end();
      }
      Files.move(temp, path, StandardCopyOption.REPLACE_EXISTING);
    } catch (IOException e) {
      if (temp != null) try {
        Files.deleteIfExists(temp);
      } catch (IOException _e) {
        // nothing to do
      }
      Logger.warning("Failed to save callgraph to '%s': %s", path, e.getMessage());
    }
  }

  private static final class DotWriter implements AutoCloseable {
    final PrintWriter writer;
    final Map<String, String> nodeNames = new HashMap<>();

    DotWriter(Path path) throws IOException {
      writer = new PrintWriter(Files.newBufferedWriter(path));
    }

    void start(String name) {
      writer.printf("digraph %s {%n", name);
    }

    void end() {
      writer.println("}");
    }

    void writeAll(String label, Collection<String> targets) {
      if (targets.isEmpty()) {
        return;
      }
      final String source = nameOf(label);
      for (String tlabel: targets) {
        final String target = nameOf(tlabel);
        writer.printf("  %s -> %s%n", source, target);
      }
    }

    String nameOf(String label) {
      return nodeNames.computeIfAbsent(label, this::registerLabel);
    }

    String registerLabel(String label) {
      final var name = String.format("node%d", nodeNames.size());
      writer.printf("  %s[label=\"%s\"];%n", name, label);
      return name;
    }

    @Override
    public void close() {
      writer.close();
    }
  }

  private static final class CallVisitor extends InsnVisitor {
    private Set<String> callees = Set.of();

    @Override
    public void visit(Instruction.Call call) {
      callees.add(call.method);
    }
  }
}
