/*
 *  Copyright 2021 ISP RAS (http://www.ispras.ru)
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package ru.ispras.microtesk.translator.mir;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import ru.ispras.fortress.data.Data;
import ru.ispras.fortress.data.DataType;
import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.types.TypeVariant;
import ru.ispras.fortress.data.types.UnionType;
import ru.ispras.fortress.data.types.datamap.DataMap;
import ru.ispras.fortress.expression.Node;
import ru.ispras.fortress.expression.NodeBinding;
import ru.ispras.fortress.expression.NodeOperation;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.expression.NodeVariable;
import ru.ispras.fortress.expression.Nodes;
import ru.ispras.fortress.solver.engine.smt.AssertionLevel;
import ru.ispras.fortress.solver.engine.smt.Cvc4Solver;
import ru.ispras.fortress.solver.engine.smt.SmtTextBuilder;
import ru.ispras.fortress.solver.function.Function;
import ru.ispras.fortress.solver.function.FunctionOp;
import ru.ispras.microtesk.translator.mir.Mir2Node.Opnd2Node;
import ru.ispras.microtesk.translator.mir.util.NmlInstances;
import ru.ispras.microtesk.translator.mir.util.PathAnalysis;
import ru.ispras.microtesk.translator.mir.util.PathAnalysis.PathConditionWalkAction;
import ru.ispras.microtesk.translator.mir.util.SmtLib;
import ru.ispras.microtesk.translator.nml.ir.Ir;
import ru.ispras.microtesk.utils.NamePath;

/** Translator of nML instances to SMT-Lib. 
 */
public class NmlToSmtTranslator {
  private final Path outdir;
  private ArchiveWriter archive = null;

  public NmlToSmtTranslator(Path outdir) {
    this.outdir = outdir;
  }

  /** Translate instances of the given {@link Ir} to SMT-Lib and store its
   *  path conditions as constraints into external archive.
   *
   * @param ir the IR to translate from
   *
   * @param compiled MIR representation of the IR
   *
   * @param funcs MIR representation of recursive functions in IR
   */
  public void run(Ir ir, Map<String, MirContext> compiled, Map<String, MirContext> funcs) {
    final List<MirContext> instances = NmlInstances.produce(ir, compiled);
    final var binders = NmlInstances.produceBinders(ir, compiled);

    final var smt = new SmtLib();
    final UnionType stateSort = smt.declareStateSort(ir, "State");
    final TypeVariant stateTy = stateSort.listVariants().get(0);

    final Map<String, Function> funcsSmt = translateFunctions(stateTy, funcs);
    final Path path = outdir.resolve(ir.getModelName() + "-paths.zip");
    try {
      Files.createDirectories(path.getParent());
      try (final var aw = new ArchiveWriter(path)) {
        this.archive = aw;
        for (int i = 0; i < instances.size(); ++i) {
          processMir(stateTy, instances.get(i), binders.get(i), funcsSmt);
        }
      }
    } catch (IOException e) {
      throw new UncheckedIOException(e);
    }
  }

  public static Map<String, Function> translateFunctions(
      TypeVariant stateTy, Map<String, MirContext> mirs) {
    final NodeVariable statePar = stateTy.getType().newNodeVar("%state");
    final var funcs = new TreeMap<String, Function>();
    for (final MirContext mir : mirs.values()) {
      final List<NodeVariable> parList =
          FunctionBuilder.parametersOf(statePar, mir);
      final DataType valueType = Mir2Node.typeOf(mir.getSignature().ret);
      final var f = new Function(mir.name, parList, valueType);
      funcs.put(f.getUniqueName(), f);
    }

    final var driver = new MirPassDriver(new GlobalNumbering());
    for (final MirContext mir : mirs.values()) {
      final var ssa = driver.apply(mir);
      final var compiled = new FunctionBuilder(stateTy, ssa, funcs);
      final Node body = compiled.buildFunctionBody(compiled.output);

      final Function f = funcs.get(mir.name);
      f.setBody(body);
    }
    return funcs;
  }

  private void processMir(
      TypeVariant stateTy,
      MirContext mir,
      NmlInstances.BinderInfo binder,
      Map<String, Function> funcs) throws IOException {
    final var compiled = new FunctionBuilder(stateTy, mir, funcs);
    /* NOTE An example of how to construct an actual instruction body in SMT,
     * presented here for reference
    final Function bodyFn = compiled.buildFunction(mir.name, compiled.output);
    */

    final var action = new ConditionAction(mir.name, compiled);
    PathAnalysis.analysePaths(mir, action);

    final Function bindFn = compiled.buildBinder(mir.name + "!bind", binder.mir);

    final List<Node> bindArgs = new ArrayList<>();
    final List<Node> pathArgs = new ArrayList<>();
    bindArgs.add(compiled.stateArg);
    pathArgs.add(compiled.stateArg);

    final Map<NamePath, NodeVariable> argMap = new LinkedHashMap<>();
    for (final var entry : binder.inputAlias.entrySet()) {
      final NamePath name = entry.getKey();
      final DataType type = Mir2Node.typeOf(entry.getValue());
      final var arg = new NodeVariable(name.toString(), type);
      argMap.put(name, arg);
      bindArgs.add(arg);
    }
    for (final NamePath parName : binder.pathInput) {
      final var arg = argMap.get(parName);
      pathArgs.add(arg);
    }

    final Node assertBind = applyFn(bindFn, bindArgs);
    final var level = AssertionLevel.of(List.of(assertBind));
    for (Function fn : action.conditions) {
      final Node assertPath = applyFn(fn, pathArgs);
      level.addAssertion(List.of(assertPath));
    }

    final var solver = new Cvc4Solver();
    final String name = mir.name + ".smt2";
    try (final var writer = archive.newText(name)) {
      SmtTextBuilder.writeTo(
          writer, List.of(), level, solver.getOperations());
    }
  }

  private static NamePath prefixOf(NamePath name) {
    return name.subpath(0, name.getNameCount() - 1);
  }

  private static NodeOperation applyFn(Function fn, List<Node> args) {
    return NodeOperation.of(FunctionOp.APPLY, List.of(fn), args);
  }

  private static final class ConditionAction extends PathConditionWalkAction {
    final String prefix;
    final FunctionBuilder compiled;
    final List<Function> conditions = new ArrayList<>();
    final OperandWalker<Node> walker =
        new OperandWalker<>(new Opnd2Node(Map.of(), new TreeMap<>()));

    ConditionAction(String prefix, FunctionBuilder compiled) {
      this.prefix = prefix;
      this.compiled = compiled;
    }

    @Override
    public void accept(List<Operand> pos, List<Operand> neg) {
      final List<Node> args = new ArrayList<>();
      for (final Node node : walker.dispatchAll(pos)) {
        args.add(Nodes.eq(node, Mir2Node.BIT_ONE));
      }
      for (final Node node : walker.dispatchAll(neg)) {
        args.add(Nodes.eq(node, Mir2Node.BIT_ZERO));
      }
      final var name = String.format("%s!%d", prefix, conditions.size());
      final Function fn = compiled.buildFunction(name, Nodes.and(args));
      conditions.add(fn);
    }
  }

  public static final class FunctionBuilder {
    final TypeVariant stateTy;
    final NodeVariable stateArg;
    final List<NodeVariable> parList;
    final NodeBinding input;
    final NodeBinding defaulted;
    final Node output;
    final Mir2Node.AssignList assignList;

    public  FunctionBuilder(
        TypeVariant stateTy, MirContext mir, Map<String, Function> funcs) {
      this(stateTy, mir, funcs, Map.of());
    }

    public FunctionBuilder(
        TypeVariant stateTy,
        MirContext mir,
        Map<String, Function> funcs,
        Map<String, Node> overrides) {
      this.stateTy = stateTy;

      final var analysis = new StoreAnalysis();
      analysis.apply(mir);

      stateArg = stateTy.getType().newNodeVar("%state");

      parList = parametersOf(stateArg, mir);

      final var visitor = new StateAwareVisitor(stateArg, funcs);
      final var compiler = new Mir2Node(visitor);
      compiler.apply(mir);

      input = bindInput(stateArg, stateTy, analysis.versionMap());
      output = visitor.hasReturnValue()
          ? visitor.getReturnValue()
          : buildStateOutput(stateArg, stateTy, analysis.modifiedMap(), overrides);

      assignList = compiler.getAssignList();

      // 'toNested()' call is required to populate free vars
      assignList.toNested(output);
      defaulted = bindFreeVariables(input, parList, assignList.getFreeVars());
    }

    private static List<NodeVariable> parametersOf(NodeVariable stateArg, final MirContext mir) {
      final List<NodeVariable> parList = new ArrayList<>();
      parList.add(stateArg);

      final List<MirTy> parTyList = mir.getSignature().params;
      for (int i = 0; i < parTyList.size(); ++i) {
        final DataType type = Mir2Node.typeOf(parTyList.get(i));
        final String name = String.format("%%%d", i + 1);
        parList.add(new NodeVariable(name, type));
      }
      return parList;
    }

    public Function buildFunctionOutput(String name) {
      return buildFunction(name, this.output);
    }

    Function buildFunction(String name, Node ret) {
      final Node body = buildFunctionBody(ret);
      return new Function(name, this.parList, body);
    }

    Node buildFunctionBody(Node ret) {
      final Node body = assignList.toNested(ret);
      final Node expr =
          defaulted.getBindings().isEmpty() ? body : defaulted.bindTo(body);
      return this.input.bindTo(expr);
    }

    Function buildBinder(String name, MirContext mir) {
      final var binderPar = parametersOf(stateArg, mir);

      final var analysis = new StoreAnalysis();
      analysis.apply(mir);
      final NodeBinding binderInput = bindInput(stateArg, stateTy, analysis.versionMap());

      final var compiler = new Mir2Node();
      compiler.apply(mir);

      final List<Node> nodeList = compiler.getAssignList().toList();
      final List<NodeBinding.BoundVariable> boundList = new ArrayList<>();

      final var it = nodeList.iterator();
      while (it.hasNext()) {
        final var op = (NodeOperation) it.next();
        final var lhs = (NodeVariable) op.getOperand(0);
        if (!binderPar.contains(lhs)) {
          final var bound = NodeBinding.bindVariable(lhs, op.getOperand(1));
          boundList.add(bound);

          it.remove();
        }
      }
      if (nodeList.isEmpty()) {
        return new Function(name, binderPar, Nodes.TRUE);
      }
      Collections.reverse(boundList);
      Node body = Nodes.and(nodeList);
      for (final NodeBinding.BoundVariable bound : boundList) {
        body = new NodeBinding(body, List.of(bound));
      }
      if (!binderInput.getBindings().isEmpty()) {
        body = binderInput.bindTo(body);
      }
      return new Function(name, binderPar, body);
    }
  }

  private static NodeBinding bindFreeVariables(
      NodeBinding input, List<NodeVariable> parList, Collection<NodeVariable> freeVars) {
    freeVars.removeAll(parList);
    for (final var bound : input.getBindings()) {
      freeVars.remove(bound.getVariable());
    }
    return bindDefaultValues(input, freeVars);
  }

  private static Node buildStateOutput(
      Node stateArg, TypeVariant stateTy, Map<String, Static> refs, Map<String, Node> overrides) {
    final Map<String, Node> values = new TreeMap<>();
    for (final var entry : stateTy.getFields().entrySet()) {
      final String name = entry.getKey();
      if (refs.containsKey(name)) {
        final Static output = refs.get(name);
        final Node value = entry.getValue().newNodeVar(output.toString());
        values.put(name, value);
      } else {
        values.put(name, stateTy.newSelector(name, stateArg));
      }
    }
    values.putAll(overrides);

    return stateTy.newConstruct(values);
  }

  private static NodeBinding bindInput(
      Node stateArg, TypeVariant stateTy, Map<String, Static> refs) {
    final List<NodeBinding.BoundVariable> bound = new ArrayList<>();
    for (final String name : stateTy.getFields().keySet()) {
      if (refs.containsKey(name)) {
        final Node select = stateTy.newSelector(name, stateArg);
        final Static input = refs.get(name).newVersion(1);
        final var boundVar = bindName(input.toString(), select);
        bound.add(boundVar);
      }
    }
    return new NodeBinding(stateArg, bound);
  }

  private static NodeBinding.BoundVariable bindName(String name, Node expr) {
    final var v = new NodeVariable(name, expr.getDataType());
    return NodeBinding.bindVariable(v, expr);
  }

  private static NodeBinding bindDefaultValues(
      final Node expr, final Collection<NodeVariable> vars) {
    final List<NodeBinding.BoundVariable> bound = new ArrayList<>(vars.size());
    for (final var v : vars) {
      final var value = new NodeValue(defaultValue(v.getDataType()));
      bound.add(NodeBinding.bindVariable(v, value));
    }
    return new NodeBinding(expr, bound);
  }

  private static Data defaultValue(final DataType type) {
    final var typeId = type.getTypeId();
    switch (typeId) {
    default:
      throw new UnsupportedOperationException();

    case BIT_VECTOR:
      return Data.newBitVector(0, type.getSize());

    case MAP:
      final var keyT = (DataType) type.getAttribute(DataTypeId.Attribute.KEY);
      final var valT = (DataType) type.getAttribute(DataTypeId.Attribute.VALUE);
      final var map = new DataMap(keyT, valT);
      map.setConstant(defaultValue(valT));

      return Data.newMap(map);
    }
  }

  private static final class StateAwareVisitor extends Mir2Node.InsnToSmtVisitor {
    private final NodeVariable stateArg;
    private final Map<String, Function> funcs;
    private Node retValue;

    StateAwareVisitor(NodeVariable stateArg, Map<String, Function> funcs) {
      this.stateArg = stateArg;
      this.funcs = funcs;
    }

    boolean hasReturnValue() {
      return retValue != null;
    }

    Node getReturnValue() {
      return retValue;
    }

    @Override
    public void visit(Instruction.Call call) {
      final List<Node> args = dispatchAll(call.args);
      args.add(0, this.stateArg);

      final Function f = funcs.get(call.method);
      assign(call.ret, applyFn(f, args));
    }

    @Override
    public void visit(Instruction.Return insn) {
      if (insn.value != null) {
        retValue = dispatch(insn.value);
      }
    }
  }
}
