/*
    Copyright 2023 ISP RAS (http://www.ispras.ru)

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
*/

package ru.ispras.microtesk.translator.mir;

import static ru.ispras.microtesk.translator.mir.GlobalNumbering.SsaStore;
import static ru.ispras.microtesk.translator.mir.GlobalNumbering.Phi;
import static ru.ispras.microtesk.translator.mir.Instruction.Load;
import static ru.ispras.microtesk.translator.mir.Instruction.Store;

public class SsaForwardPass extends ForwardPass {
  @Override
  public MirContext apply(MirContext source) {
    final MirContext ctx = Pass.copyOf(source);

    insertDefinitionPoints(ctx);
    InsnRewriter.rewrite(ctx, this.presets);

    return ctx;
  }

  private static void insertDefinitionPoints(MirContext mir) {
    for (BasicBlock bb: mir.blocks) {
      for (int i = 0; i < bb.insns.size(); ++i) {
        final Static source = getScalarTarget(bb.insns.get(i));
        if (source != null) {
          final var target = newLocal(source.getType(), bb.getOrigin(i), mir);
          final var load = new Load(source, target);
          bb.insns.add(i + 1, load);
          shiftOrigins(bb, i + 1, 1);
        }
      }
    }
  }

  private static void shiftOrigins(BasicBlock bb, int index, int offset) {
    for (var org: bb.origins) {
      if (org.range >= index) {
        org.range += offset;
      }
    }
  }

  private static Local newLocal(MirTy type, int org, MirContext mir) {
    mir.locals.add(type);
    return new Local(mir.locals.size() - 1 - org, type);
  }

  private static Static getScalarTarget(Instruction insn) {
    Lvalue target = null;
    if (insn instanceof Store) {
      target = ((Store) insn).target;
    } else if (insn instanceof SsaStore) {
      target = ((SsaStore) insn).target;
    } else if (insn instanceof Phi) {
      target = ((Phi) insn).target;
    }
    if (target instanceof Static && !Types.isArray(target)) {
      return (Static) target;
    }
    return null;
  }
}
