/*
 *  Copyright 2021 ISP RAS (http://www.ispras.ru)
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package ru.ispras.microtesk.translator.mir.util;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import ru.ispras.castle.util.Logger;
import ru.ispras.microtesk.translator.mir.Lists;
import ru.ispras.microtesk.translator.mir.MirBuilder;
import ru.ispras.microtesk.translator.mir.MirContext;
import ru.ispras.microtesk.translator.mir.MirPassDriver;
import ru.ispras.microtesk.translator.mir.MirTy;
import ru.ispras.microtesk.translator.mir.NmlIrTrans;
import ru.ispras.microtesk.translator.nml.ir.Ir;
import ru.ispras.microtesk.translator.nml.ir.primitive.Primitive;
import ru.ispras.microtesk.translator.nml.ir.primitive.PrimitiveAnd;
import ru.ispras.microtesk.translator.nml.ir.primitive.PrimitiveOr;
import ru.ispras.microtesk.utils.NamePath;

/** Contains utilities to construct complete instances of nML primitives.
 */
public final class NmlInstances {
  public static List<MirContext> produce(final Ir ir, final Map<String, MirContext> source) {
    final MirPassDriver driver = MirPassDriver.newOptimizing();
    driver.setStorage(source);

    final List<PrimitiveAnd> opList = ir.getOps().containsKey("instruction")
        ? variantsOf(ir.getOps().get("instruction"))
        : Collections.emptyList();

    final ProcessReport report = new ProcessReport();
    final List<MirContext> isa = opList.stream()
        .flatMap(x -> instancesOf(x, report))
        .map(x -> report.notifyDone(driver.apply(x)))
        .collect(Collectors.toList());
    return isa;
  }

  public static List<BinderInfo> produceBinders(
      final Ir ir, final Map<String, MirContext> source) {
    final MirPassDriver driver = MirPassDriver.newOptimizing();
    driver.setStorage(source);

    final List<PrimitiveAnd> opList = ir.getOps().containsKey("instruction")
        ? variantsOf(ir.getOps().get("instruction"))
        : Collections.emptyList();

    final ProcessReport report = new ProcessReport();
    final List<BinderInfo> binders = opList.stream()
        .flatMap(x -> bindersOf(x, report))
        .collect(Collectors.toList());
    for (var binder : binders) {
      binder.mir = driver.apply(binder.builder.build());
    }
    return binders;
  }

  static class ProcessReport {
    private int total = 0;
    private int ndone = 0;
    private int cycle = 0;

    <T> Collection<T> addAll(Collection<T> c) {
      total += c.size();
      return c;
    }

    <T> T notifyDone(T item) {
      final int threshold = Math.min(500, Math.max(total / 5 - 1, 1));
      ndone += 1;
      cycle += 1;
      if (cycle > threshold) {
        Logger.message("  %d/%d...", ndone, total);
        cycle %= threshold;
      }
      return item;
    }
  }

  static Stream<MirContext> instancesOf(final PrimitiveAnd p, final ProcessReport report) {
    final var name = p.getName() + ".action";
    final var instances = instantiateRec(p, new MirBuilder(p.getName()));
    report.addAll(instances);

    return instances.stream()
      .peek(b -> b.makeCall(name, 0))
      .map(MirBuilder::build);
  }

  private static List<MirBuilder> instantiateRec(
      final PrimitiveAnd op, final MirBuilder builder) {
    final List<MirBuilder> queue = Lists.newList(builder);
    final List<MirBuilder> swap = Lists.newList();
    for (final var entry : op.getArguments().entrySet()) {
      final Primitive p = entry.getValue();
      if (p.getKind().equals(Primitive.Kind.IMM)) {
        for (final var b : queue) {
          b.refParameter(b.addParameter(NmlIrTrans.typeOf(p)));
        }
      } else {
        for (final PrimitiveAnd v : variantsOf(p)) {
          for (final var b : queue) {
            final String name = b.getName() + "_" + v.getName();
            swap.addAll(instantiateRec(v, b.copyAs(name)));
          }
        }
        Lists.moveAll(queue, swap);
      }
    }
    for (final var b : queue) {
      b.makeClosure(op.getName(), op.getArguments().size());
    }
    return queue;
  }

  public static final class BinderInfo {
    final MirBuilder builder;
    public final Map<NamePath, MirTy> inputAlias;
    public final List<NamePath> pathInput;
    public MirContext mir = null;

    BinderInfo(String name) {
      this.builder = new MirBuilder(name);
      this.inputAlias = new LinkedHashMap<>();
      this.pathInput = new ArrayList<>();
    }

    private BinderInfo(String name, BinderInfo that) {
      this.builder = that.builder.copyAs(name);
      this.inputAlias = new LinkedHashMap<>(that.inputAlias);
      this.pathInput = new ArrayList<>(that.pathInput);
    }

    BinderInfo copyAs(String name) {
      return new BinderInfo(name, this);
    }
  }

  static Stream<BinderInfo> bindersOf(final PrimitiveAnd p, final ProcessReport report) {
    final String name = p.getName();
    final var instances = instantiateBinderRec(
        p, NamePath.get(name), new BinderInfo(name));
    return instances.stream();
  }

  private static List<BinderInfo> instantiateBinderRec(
      final PrimitiveAnd op, NamePath namePrefix, final BinderInfo info) {
    final List<BinderInfo> queue = Lists.newList(info);
    final List<BinderInfo> swap = Lists.newList();
    for (final var entry : op.getArguments().entrySet()) {
      final Primitive p = entry.getValue();
      final NamePath nestedName = namePrefix.resolve(entry.getKey());
      if (p.getKind().equals(Primitive.Kind.IMM)) {
        for (final var b : queue) {
          final MirTy ty = NmlIrTrans.typeOf(p);
          b.inputAlias.put(nestedName, ty);
          b.pathInput.add(nestedName);

          final int pid = b.builder.addParameter(ty);
          if (op.getKind().equals(Primitive.Kind.MODE)) {
            b.builder.refParameter(pid);
          }
        }
      } else if (p.isOrRule()) {
        for (final PrimitiveAnd v : variantsOf(p)) {
          for (final var b : queue) {
            final String name = b.builder.getName() + "_" + v.getName();
            swap.addAll(instantiateBinderRec(v, nestedName, b.copyAs(name)));
          }
        }
        Lists.moveAll(queue, swap);
      } else {
        for (final var b : queue) {
          final var rule = (PrimitiveAnd) p;
          swap.addAll(instantiateBinderRec(rule, nestedName, b));
        }
        Lists.moveAll(queue, swap);
      }
    }
    if (op.getKind().equals(Primitive.Kind.MODE)) {
      for (final var item : queue) {
        final MirBuilder b = item.builder;

        final MirTy ty = NmlIrTrans.returnTypeOf(op);
        item.inputAlias.put(namePrefix, ty);

        final int pid = b.addParameter(ty);
        b.makeClosure(op.getName(), op.getArguments().size());
        b.refParameter(pid);
        b.makeCallRet(op.getName() + ".read", 0);
      }
    }
    return queue;
  }

  static List<PrimitiveAnd> variantsOf(final Primitive p) {
    if (p.isOrRule()) {
      final List<PrimitiveAnd> variants = new java.util.ArrayList<>();
      collectVariants((PrimitiveOr) p, variants);
      return variants;
    }
    return List.of((PrimitiveAnd) p);
  }

  static void collectVariants(final PrimitiveOr input, final List<PrimitiveAnd> variants) {
    for (final Primitive p : input.getOrs()) {
      if (p.isOrRule()) {
        collectVariants((PrimitiveOr) p, variants);
      } else {
        variants.add((PrimitiveAnd) p);
      }
    }
  }
}
