/*
 *  Copyright 2021 ISP RAS (http://www.ispras.ru)
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package ru.ispras.microtesk.translator.mir.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import ru.ispras.microtesk.translator.mir.BasicBlock;
import ru.ispras.microtesk.translator.mir.Instruction.Branch;
import ru.ispras.microtesk.translator.mir.Lists;
import ru.ispras.microtesk.translator.mir.MirContext;
import ru.ispras.microtesk.translator.mir.Operand;

/** Contains utilities to analyse syntactically available paths and construct
 *  corresponding path conditions.
 */
public class PathAnalysis {
  public static void analysePaths(final MirContext mir, final PathConditionWalkAction action) {
    final GuardNode entry = ConditionGraphConstructAction.run(mir);
    if (entry != null) {
      action.step(entry);
    }
  }

  /** An abstract class to construct and process path conditions of the given
   *  condition graph.
   */
  public static abstract class PathConditionWalkAction {
    private final List<Operand> positive = new ArrayList<>();
    private final List<Operand> negative = new ArrayList<>();

    private final void step(GuardNode node) {
      halfStep(node.guard, node.taken, positive);
      halfStep(node.guard, node.other, negative);
    }

    private final void halfStep(Operand guard, GuardNode next, List<Operand> signedList) {
      signedList.add(guard);
      if (next != null) {
        step(next);
      } else {
        accept(positive, negative);
      }
      Lists.removeLast(signedList);
    }

    /** Process a single path condition represented by boolean assignment.
     *
     * @param positive Guard conditions set to {@code true}
     * @param negative Guard conditions set to {@code false}
     */
    protected abstract void accept(List<Operand> positive, List<Operand> negative);
  }

  private static final class GuardNode {
    final Operand guard;
    GuardNode taken;
    GuardNode other;

    private GuardNode(Operand guard) {
      this.guard = guard;
    }

    static GuardNode of(Operand guard) {
      return new GuardNode(guard);
    }
  }

  /** Compress CFG into condition graph representing all syntacically
   *  available paths and path conditions
   */
  private static final class ConditionGraphConstructAction {
    final Map<String, GuardNode> nodes = new HashMap<>();

    public static GuardNode run(MirContext mir) {
      final var action = new ConditionGraphConstructAction();
      return action.step(mir.blocks.get(0));
    }

    /** Recursive depth-first walk over basic blocks 
     */
    private GuardNode step(BasicBlock bb) {
      final Branch br = branchOf(bb);
      if (br == null) {
        return null;
      }

      var succ = br.successors;
      if (succ.size() == 1) {
        return step(succ.get(0));
      }

      final String key = br.guard.toString();
      GuardNode node = nodes.get(key);
      if (node != null) {
        return node;
      }

      node = GuardNode.of(br.guard);
      nodes.put(key, node);

      node.taken = step(succ.get(0));
      node.other = step(succ.get(1));

      return node;
    }
  }

  private static Branch branchOf(BasicBlock bb) {
    var insn = Lists.lastOf(bb.insns);
    if (insn instanceof Branch) {
      return (Branch) insn;
    }
    return null;
  }
}
