/*
 *  Copyright 2021 ISP RAS (http://www.ispras.ru)
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */

package ru.ispras.microtesk.translator.mir.util;

import java.math.BigInteger;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import ru.ispras.fortress.data.types.MonomorphicType;
import ru.ispras.fortress.data.types.PolymorphicType;
import ru.ispras.fortress.data.types.TypeVar;
import ru.ispras.fortress.data.types.Types;
import ru.ispras.fortress.data.types.UnionType;
import ru.ispras.microtesk.model.memory.Memory;
import ru.ispras.microtesk.translator.nml.ir.Ir;
import ru.ispras.microtesk.translator.nml.ir.shared.MemoryResource;

/** Contains utilties for translation to SMT-Lib.
 */
public final class SmtLib {
  private final Types types = new Types();

  public UnionType declareStateSort(Ir ir, String sortName) {
    final var fields = new TreeMap<String, MonomorphicType>();
    for (final MemoryResource mem : ir.getMemory().values()) {
      if (!mem.getKind().equals(Memory.Kind.VAR)) {
        final var type = typeOfMemoryResource(mem);
        fields.put(mem.getName(), type);
      }
    }
    final Types.Builder builder = types.newSort(sortName, 0);
    builder.addVariant(sortName, fields);

    return builder.build();
  }

  private MonomorphicType typeOfMemoryResource(MemoryResource mem) {
    final var type = newBitVec(mem.getType().getBitSize());
    final BigInteger size = mem.getSize();
    if (size.compareTo(BigInteger.ONE) > 0) {
      final int nbits = size.subtract(BigInteger.ONE).bitLength();
      final var keyty = newBitVec(nbits);

      return newArray(keyty, type);
    }
    return type;
  }

  private MonomorphicType newBitVec(int nbits) {
    return types.getDependent("BitVec", nbits);
  }

  private MonomorphicType newArray(MonomorphicType keyTy, MonomorphicType valTy) {
    final var array = (PolymorphicType) types.get("Array");
    final List<TypeVar> par = array.getParameters();
    return array.define(Map.of(par.get(0), keyTy, par.get(1), valTy));
  }
}
