/*
 *  Copyright 2021 ISP RAS (http://www.ispras.ru)
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 */
package ru.ispras.microtesk.translator.nml.coverage;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.UncheckedIOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.StandardCopyOption;
import java.util.stream.Collectors;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import ru.ispras.fortress.data.DataTypeId;
import ru.ispras.fortress.data.Variable;
import ru.ispras.fortress.expression.NodeValue;
import ru.ispras.fortress.solver.Solver;
import ru.ispras.fortress.solver.SolverId;
import ru.ispras.fortress.solver.SolverResult;
import ru.ispras.fortress.solver.engine.smt.SmtTextSolver;
import ru.ispras.microtesk.SysUtils;
import ru.ispras.testbase.TestBaseContext;
import ru.ispras.testbase.TestBaseQuery;
import ru.ispras.testbase.TestData;
import ru.ispras.testbase.generator.DataGenerator;
import ru.ispras.testbase.generator.DataGeneratorQuery;
import ru.ispras.testbase.generator.DataGeneratorVariable;
import ru.ispras.testbase.knowledge.iterator.CollectionIterator;
import ru.ispras.testbase.knowledge.iterator.EmptyIterator;
import ru.ispras.testbase.knowledge.iterator.Iterator;

public final class GeneratorAllPaths implements DataGenerator {
  public static final String NAME = "all_paths";

  private final Map<String, ZipFile> archives = new HashMap<>();

  public String getName() {
    return NAME;
  }

  @Override
  public boolean isSuitable(TestBaseQuery query) {
    final var request = query.getContext().get(TestBaseContext.TESTCASE);
    return NAME.equals(request);
  }

  @Override
  public Iterator<TestData> generate(final TestBaseQuery sourceQuery) {
    final var query = new DataGeneratorQuery(sourceQuery);
    final Solver solver = SolverId.CVC4_TEXT.getSolver();
    if (solver instanceof SmtTextSolver) {
      final var smtSolver = (SmtTextSolver) solver;
      final Path path = extractSmtScript(query);

      final Map<String, String> suffixes = buildSuffixMap(query);
      final java.util.Iterator<SolverResult> smtIt = smtSolver.solveAll(path.toString());

      final var assignList = new LinkedHashSet<Map<String, Object>>();
      while (smtIt.hasNext()) {
        final var result = smtIt.next();
        if (result.getStatus() == SolverResult.Status.SAT) {
          final Map<String, Object> bindings = new TreeMap<>();
          for (final Variable v : result.getVariables()) {
            for (final var entry : suffixes.entrySet()) {
              if (v.getName().endsWith(entry.getKey())) {
                final var node = new NodeValue(v.getData());
                bindings.put(entry.getValue(), node);
              }
            }
          }
          assignList.add(bindings);
        }
      }
      final List<TestData> dataset = assignList.stream()
        .map(TestData::new)
        .collect(Collectors.toList());

      return new CollectionIterator<>(dataset);
    }
    return EmptyIterator.get();
  }

  private static Map<String, String> buildSuffixMap(DataGeneratorQuery query) {
    final Map<String, String> suffixes = new TreeMap<>();
    final String prefix = query.getInstructionName() + ".";
    for (final DataGeneratorVariable v : query.getUnknownVariables()) {
      final String name = v.getName();
      if (v.isType(DataTypeId.BIT_VECTOR) && name.startsWith(prefix)) {
        final String suffix = name.substring(prefix.length() - 1);
        suffixes.put(suffix, name);
      } else {
        throw new IllegalArgumentException(String.format(
            "Variable %s has unsupported type %s.", v.getName(), v.getType()));
      }
    }
    return suffixes;
  }

  private Path extractSmtScript(DataGeneratorQuery query) {
    final String modelName = query.getArchitectureName();
    final ZipFile archive =
        archives.computeIfAbsent(modelName, GeneratorAllPaths::openArchive);
    final String name = getEntryName(query);
    final ZipEntry entry = archive.getEntry(name);
    try (InputStream in = archive.getInputStream(entry)) {
      final File temp = File.createTempFile(NAME, ".smt2");
      temp.deleteOnExit();
      Files.copy(in, temp.toPath(), StandardCopyOption.REPLACE_EXISTING);
      return temp.toPath();
    } catch (IOException e) {
      throw new UncheckedIOException(e);
    }
  }

  private static String getEntryName(DataGeneratorQuery query) {
    final String name = query.getInstructionName();
    final var sb = new StringBuilder(name);
    buildInstanceName(sb, name, query.getContext());

    sb.append(".smt2");
    sb.insert(0, "instruction_");
    return sb.toString();
  }

  private static void buildInstanceName(
      StringBuilder acc, String prefix, Map<String, Object> ctx) {
    for (final String key : ctx.keySet()) {
      if (key.startsWith(prefix) && key.lastIndexOf('.') == prefix.length()) {
        final var type = ctx.get(key).toString();
        if (!type.equals("#IMM")) {
          acc.append("_").append(type);
          buildInstanceName(acc, key, ctx);
        }
      }
    }
  }

  private static ZipFile openArchive(String name) {
    try {
      final Path path =
          Paths.get(SysUtils.getHomeDir(), "gen", name + "-paths.zip");
      return new ZipFile(path.toFile());
    } catch (IOException e) {
      throw new UncheckedIOException(e);
    }
  }
}
