/*
    Copyright 2022 ISP RAS (http://www.ispras.ru)

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
*/
package ru.ispras.microtesk.utils;

import ru.ispras.fortress.data.types.bitvector.BitVector;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;

public class BitStream {
  public static BitStream of(InputStream input) {
    return new BitStream(input, 16);
  }

  public boolean isEmpty() throws IOException {
    if (offsetLo < offsetHi) {
      return false;
    }
    ensureCapacity(BYTE_SIZE);
    return advance(1) != 1;
  }

  public BitVector read(int nbits) throws IOException {
    ensureCapacity(nbits);

    final int nleft = offsetHi - offsetLo;
    if (nbits > nleft) {
      final int nbytes = nearestByteSize(nbits - nleft);
      if (advance(nbytes) < nbytes) {
        throw new java.io.EOFException();
      }
    }

    final BitVector bv = extractRange(nbits);
    offsetLo += nbits;

    return bv;
  }

  private int advance(int nbytes) throws IOException {
      final int offset = nearestByteSize(offsetHi);
      int nread = 0;
      while (nread < nbytes) {
        int rc = input.read(buffer, offset + nread, nbytes - nread);
        if (rc == -1) {
          break;
        }
        nread += rc;
      }
      offsetHi += nread * BYTE_SIZE;

      return nread;
  }

  private void ensureCapacity(int nbits) {
    final int capacity = buffer.length * BYTE_SIZE;
    final int nleft = offsetHi - offsetLo;
    final int nfree = capacity - offsetHi;

    final int nreq = nbits + offsetLo % BYTE_SIZE;
    if (nreq > capacity) {
      final var oldbuf = buffer;
      buffer = new byte[nearestByteSize(nreq)];
      shiftBytesLeft(oldbuf, offsetLo, offsetHi);
    } else if (nbits > nleft + nfree) {
      shiftBytesLeft(buffer, offsetLo, offsetHi);
    }
  }

  private void shiftBytesLeft(byte[] source, int loBits, int hiBits) {
    final int lo = loBits / BYTE_SIZE;
    final int len = nearestByteSize(hiBits) - lo;
    System.arraycopy(source, lo, buffer, 0, len);

    offsetLo = loBits % BYTE_SIZE;
    offsetHi = offsetLo + (hiBits - loBits);
  }

  private BitVector extractRange(int bitSize) {
    final int byteLo = offsetLo / BYTE_SIZE;
    final int byteHi = nearestByteSize(offsetHi);

    final var bytes = Arrays.copyOfRange(buffer, byteLo, byteHi);
    final var bv = BitVector.valueOf(bytes, bytes.length * BYTE_SIZE);

    return BitVector.newMapping(bv, offsetLo % BYTE_SIZE, bitSize);
  }

  private static int nearestByteSize(int nbits) {
    return nbits / BYTE_SIZE + Math.min(nbits % BYTE_SIZE, 1);
  }

  private BitStream(InputStream input, int bufSize) {
    this.input = input;
    this.buffer = new byte[bufSize];
    this.offsetLo = 0;
    this.offsetHi = 0;
  }

  private static final int BYTE_SIZE = 8;

  private final InputStream input;
  private byte[] buffer;
  private int offsetLo;
  private int offsetHi;
}
