/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.adapter.trace;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TestName;

import ru.ispras.castle.util.Logger;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class EventRecognizerTestCase {
  @Before
  public void prepareLogger() {
    Logger.setDebug(false);
  }

  @Rule
  public TestName name = new TestName();

  private void messageAboutTest() {
    Logger.debug("___ Test method: %s", name.getMethodName());
  }

  // Memory
  private final EventTemplate memoryEventTemplate = prepareMemoryPatternElements();

  public static EventTemplate prepareMemoryPatternElements() {
    final EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    final EventElementPattern eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    final EventElementPattern eventElementRegister = new EventElementPattern(EventElementType.REGISTER, "\\s{2,}+[a-z0-9]{2,}+");
    final EventElementPattern eventElementRegisterValue = new EventElementPattern(EventElementType.REGISTER_VALUE, "[a-f0-9]{1,}+$");

    final List<EventElementPattern> eventElementTemplates = new ArrayList<>();
    eventElementTemplates.add(eventElementThreadId);
    eventElementTemplates.add(eventElementSource);
    eventElementTemplates.add(eventElementRegister);
    eventElementTemplates.add(eventElementRegisterValue);

    final Map<EventElementType, String> eventElementPartPatterns = EventRecognizerTestCase.prepareMemoryElementPartPatterns();

    final EventTemplate eventTemplate = new EventTemplate(EventType.MEMORY, eventElementTemplates, eventElementPartPatterns);
    return eventTemplate;
  }

  public static Map<EventElementType, String> prepareMemoryElementPartPatterns() {
    final Map<EventElementType, String> eventElementPartPatterns = new HashMap<>();

    eventElementPartPatterns.put(EventElementType.THREAD_IDENTIFIER, "([0-9a-f]){6,}+");
    eventElementPartPatterns.put(EventElementType.REGISTER, "[a-z0-9]{2,}+");
    eventElementPartPatterns.put(EventElementType.REGISTER_VALUE, "[a-f0-9]{1,}+");

    return eventElementPartPatterns;
  }

  @Test
  public void memoryLineRecognize() {
    this.messageAboutTest();

    final List<EventTemplate> eventMemory = new ArrayList<>();
    eventMemory.add(memoryEventTemplate);
    final EventRecognizer eventRecognizer = new EventRecognizer(eventMemory);
    Logger.debug(eventRecognizer.toString());

    final String memoryLine = "[TID 002b0a] D/interpreter:      v1.pri = (i64) 0 | (f32) 0 | (f64) 0 | (hex) 0";
    Logger.debug("Line: \"%s\"", memoryLine);

    final ConcreteEvent concreteEvent = eventRecognizer.getEvent(memoryLine);
    Logger.debug("%s\n", concreteEvent);

    assertEquals(EventType.MEMORY, concreteEvent.getEventType());
    assertEquals("002b0a", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
    assertEquals("v1", concreteEvent.getElement(EventElementType.REGISTER));
    assertEquals("0", concreteEvent.getElement(EventElementType.REGISTER_VALUE));
  }

  private final EventTemplate memoryShortEventTemplate = prepareMemoryShortPatternElements();

  public static EventTemplate prepareMemoryShortPatternElements() {
    final EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    final EventElementPattern eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    final EventElementPattern eventElementRegister = new EventElementPattern(EventElementType.REGISTER, "\\s{2,}+[a-z0-9]{2,}+");

    final List<EventElementPattern> eventElementTemplates = new ArrayList<>();
    eventElementTemplates.add(eventElementThreadId);
    eventElementTemplates.add(eventElementSource);
    eventElementTemplates.add(eventElementRegister);

    final Map<EventElementType, String> eventElementPartPatterns = EventRecognizerTestCase.prepareElementPartPatternsShort();

    final EventTemplate eventTemplate = new EventTemplate(EventType.MEMORY, eventElementTemplates, eventElementPartPatterns);
    return eventTemplate;
  }

  public static Map<EventElementType, String> prepareElementPartPatternsShort() {
    final Map<EventElementType, String> eventElementPartPatterns = new HashMap<>();

    eventElementPartPatterns.put(EventElementType.THREAD_IDENTIFIER, "([0-9a-f]){6,}+");
    eventElementPartPatterns.put(EventElementType.REGISTER, "[a-z0-9]{2,}+");

    return eventElementPartPatterns;
  }

  @Test
  public void memoryShortLineRecognize() {
    this.messageAboutTest();

    final List<EventTemplate> eventMemoryShort = new ArrayList<>();
    eventMemoryShort.add(memoryShortEventTemplate);
    final EventRecognizer eventRecognizer = new EventRecognizer(eventMemoryShort);
    Logger.debug(eventRecognizer.toString());

    final String memoryLine = "[TID 002b0a] D/interpreter:      v1.pri = (i64) 0 | (f32) 0 | (f64) 0 | (hex) 0";
    Logger.debug("Line: \"%s\"", memoryLine);

    final ConcreteEvent concreteEvent = eventRecognizer.getEvent(memoryLine);
    Logger.debug("%s\n", concreteEvent);

    assertEquals(EventType.MEMORY, concreteEvent.getEventType());
    assertEquals("002b0a", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
    assertEquals("v1", concreteEvent.getElement(EventElementType.REGISTER));
  }

  // Instructions
  private final EventTemplate instructionEventTemplate = prepareInstructionPatternElements();

  public static EventTemplate prepareInstructionPatternElements() {
    final EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    final EventElementPattern eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    final EventElementPattern eventElementPc = new EventElementPattern(EventElementType.PC, "([a-f0-9]){2,}+:");
    final EventElementPattern eventElementInstruction = new EventElementPattern(EventElementType.INSTRUCTION, "\\S([a-z0-9|\\s|,]|[.]|[->]){2,}+");

    final List<EventElementPattern> eventElementTemplates = new ArrayList<>();
    eventElementTemplates.add(eventElementThreadId);
    eventElementTemplates.add(eventElementSource);
    eventElementTemplates.add(eventElementPc);
    eventElementTemplates.add(eventElementInstruction);

    final Map<EventElementType, String> eventElementPartPatterns = EventRecognizerTestCase.prepareInstructionElementPartPatterns();

    final EventTemplate eventTemplate = new EventTemplate(EventType.INSTRUCTION, eventElementTemplates, eventElementPartPatterns);
    return eventTemplate;
  }

  public static Map<EventElementType, String> prepareInstructionElementPartPatterns() {
    final Map<EventElementType, String> eventElementPartPatterns = new HashMap<>();

    eventElementPartPatterns.put(EventElementType.THREAD_IDENTIFIER, "([0-9a-f]){6,}+");
    eventElementPartPatterns.put(EventElementType.PC, "([a-f0-9]){2,}+");
    eventElementPartPatterns.put(EventElementType.INSTRUCTION, "([a-z0-9|\\s|,]|[.]|[->]){2,}+");

    return eventElementPartPatterns;
  }

  @Test
  public void instructionLineRecognize() {
    this.messageAboutTest();

    final List<EventTemplate> eventInstruction = new ArrayList<>();
    eventInstruction.add(instructionEventTemplate);
    final EventRecognizer eventRecognizer = new EventRecognizer(eventInstruction);
    Logger.debug(eventRecognizer.toString());

    final String instructionLine = "[TID 002b0a] D/interpreter: 7f7c3f5c91b2: movi v3, 0x3";
    Logger.debug("Line: \"%s\"", instructionLine);

    final ConcreteEvent concreteEvent = eventRecognizer.getEvent(instructionLine);
    Logger.debug("%s\n", concreteEvent);

    assertEquals(EventType.INSTRUCTION, concreteEvent.getEventType());
    assertEquals("002b0a", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
    assertEquals("7f7c3f5c91b2", concreteEvent.getElement(EventElementType.PC));
    assertEquals("movi v3, 0x3", concreteEvent.getElement(EventElementType.INSTRUCTION));
  }

  // Frame creation example
  private final EventTemplate frameCreationEventTemplate = prepareFrameCreationPatternElements();

  public static EventTemplate prepareFrameCreationPatternElements() {
    final EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    final EventElementPattern eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    final EventElementPattern eventElementFrameCreation = new EventElementPattern(EventElementType.FRAME_ENTRY, "(.*?)(\\bEntry\\b)(.*)$");

    final List<EventElementPattern> eventElementTemplates = new ArrayList<>();
    eventElementTemplates.add(eventElementThreadId);
    eventElementTemplates.add(eventElementSource);
    eventElementTemplates.add(eventElementFrameCreation);

    final Map<EventElementType, String> eventElementPartPatterns = EventRecognizerTestCase.prepareFrameCreationElementPartPatterns();

    final EventTemplate eventTemplate = new EventTemplate(EventType.FRAME_ENTRY, eventElementTemplates, eventElementPartPatterns);
    return eventTemplate;
  }

  public static Map<EventElementType, String> prepareFrameCreationElementPartPatterns() {
    final Map<EventElementType, String> eventElementPartPatterns = new HashMap<>();

    eventElementPartPatterns.put(EventElementType.THREAD_IDENTIFIER, "([0-9a-f]){6,}+");
    eventElementPartPatterns.put(EventElementType.FRAME_ENTRY, "[a-zA-Z0-9]{1,}+::[a-zA-Z0-9]{1,}+");

    return eventElementPartPatterns;
  }

  @Test
  public void frameCreationLineRecognize() {
    this.messageAboutTest();

    final List<EventTemplate> eventFrameCreation = new ArrayList<>();
    eventFrameCreation.add(frameCreationEventTemplate);
    final EventRecognizer eventRecognizer = new EventRecognizer(eventFrameCreation);
    Logger.debug(eventRecognizer.toString());

    final String instructionLine = "[TID 002b0a] D/interpreter: Entry: class8060::method140";
    Logger.debug("Line: \"%s\"", instructionLine);

    final ConcreteEvent concreteEvent = eventRecognizer.getEvent(instructionLine);
    Logger.debug("%s\n", concreteEvent);

    assertEquals(EventType.FRAME_ENTRY, concreteEvent.getEventType());
    assertEquals("002b0a", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
    assertEquals("class8060::method140", concreteEvent.getElement(EventElementType.FRAME_ENTRY));
  }

  @Test
  public void frameCreationLineRecognizeGroupTrue() {
    this.messageAboutTest();

    final List<EventTemplate> eventFrameCreation = new ArrayList<>();
    eventFrameCreation.add(this.frameCreationEventTemplate);
    eventFrameCreation.add(this.memoryEventTemplate);
    eventFrameCreation.add(this.instructionEventTemplate);

    final EventRecognizer eventRecognizer = new EventRecognizer(eventFrameCreation);
    Logger.debug(eventRecognizer.toString());

    final String instructionLine = "[TID 002b0a] D/interpreter: Entry: class8060::method140";
    Logger.debug("Line: \"%s\"", instructionLine);

    final ConcreteEvent concreteEvent = eventRecognizer.getEvent(instructionLine);
    Logger.debug("%s\n", concreteEvent);

    assertEquals(EventType.FRAME_ENTRY, concreteEvent.getEventType());
    assertEquals("002b0a", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
    assertEquals("class8060::method140", concreteEvent.getElement(EventElementType.FRAME_ENTRY));
  }

  @Test
  public void frameCreationLineRecognizeGroupFalse() {
    this.messageAboutTest();

    final List<EventTemplate> eventFrameCreation = new ArrayList<>();
    eventFrameCreation.add(this.memoryEventTemplate);
    eventFrameCreation.add(this.instructionEventTemplate);
    eventFrameCreation.add(this.frameCreationEventTemplate);

    final EventRecognizer eventRecognizer = new EventRecognizer(eventFrameCreation);
    Logger.debug(eventRecognizer.toString());

    final String instructionLine = "[TID 002b0a] D/interpreter: Entry: class8060::method140";
    Logger.debug("Line: \"%s\"", instructionLine);

    final ConcreteEvent concreteEvent = eventRecognizer.getEvent(instructionLine);
    Logger.debug("%s\n", concreteEvent);

    assertNotEquals(EventType.FRAME_ENTRY, concreteEvent.getEventType());
    assertNotEquals("class8060::method140", concreteEvent.getElement(EventElementType.FRAME_ENTRY));

    assertEquals("002b0a", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
  }

  private ConcreteEvent recognizeGroupFrameMemInstr(final String instructionLine) {
    final List<EventTemplate> eventGroup = new ArrayList<>();
    eventGroup.add(this.frameCreationEventTemplate);
    eventGroup.add(this.memoryEventTemplate);
    eventGroup.add(this.instructionEventTemplate);

    final EventRecognizer eventRecognizer = new EventRecognizer(eventGroup);
    Logger.debug(eventRecognizer.toString());

    final ConcreteEvent concreteEvent = eventRecognizer.getEvent(instructionLine);
    return concreteEvent;
  }

  @Test
  public void lineRecognizeGroup00Instruction() {
    this.messageAboutTest();

    final String instructionLine = "[TID 002b0a] D/interpreter: 7f7c3f5c91ea: sta v2";
    Logger.debug("Line: \"%s\"", instructionLine);

    final ConcreteEvent concreteEvent = this.recognizeGroupFrameMemInstr(instructionLine);
    Logger.debug("%s\n", concreteEvent);

    assertEquals(EventType.INSTRUCTION, concreteEvent.getEventType());
    assertEquals("002b0a", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
    assertEquals("7f7c3f5c91ea", concreteEvent.getElement(EventElementType.PC));
    assertEquals("sta v2", concreteEvent.getElement(EventElementType.INSTRUCTION));
  }

  @Test
  public void lineRecognizeGroup01Instruction() {
    this.messageAboutTest();

    final String instructionLine = "[TID 002b0a] D/interpreter: 7f7c3f5c91ce: call.virt.short 0x0, v0, v3";
    Logger.debug("Line: \"%s\"", instructionLine);

    final ConcreteEvent concreteEvent = this.recognizeGroupFrameMemInstr(instructionLine);
    Logger.debug("%s\n", concreteEvent);

    assertEquals(EventType.INSTRUCTION, concreteEvent.getEventType());
    assertEquals("002b0a", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
    assertEquals("7f7c3f5c91ce", concreteEvent.getElement(EventElementType.PC));
    assertEquals("call.virt.short 0x0, v0, v3", concreteEvent.getElement(EventElementType.INSTRUCTION));
  }

  // Memory: registers type
  private final EventTemplate memoryTypeEventTemplate = prepareMemoryTypePatternElements();

  public static EventTemplate prepareMemoryTypePatternElements() {
    final EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    final EventElementPattern eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    final EventElementPattern eventElementRegister = new EventElementPattern(EventElementType.REGISTER, "\\s{2,}+[a-z0-9]{2,}+");
    final EventElementPattern eventElementRegisterType = new EventElementPattern(EventElementType.REGISTER_TYPE, "(obj|pri){1}");
    final EventElementPattern eventElementRegisterValue = new EventElementPattern(EventElementType.REGISTER_VALUE, "[a-f0-9]{1,}+$");

    final List<EventElementPattern> eventElementTemplates = new ArrayList<>();
    eventElementTemplates.add(eventElementThreadId);
    eventElementTemplates.add(eventElementSource);
    eventElementTemplates.add(eventElementRegister);
    eventElementTemplates.add(eventElementRegisterType);
    eventElementTemplates.add(eventElementRegisterValue);

    final Map<EventElementType, String> eventElementPartPatterns = EventRecognizerTestCase.prepareMemoryTypeElementPartPatterns();

    final EventTemplate eventTemplate = new EventTemplate(EventType.MEMORY, eventElementTemplates, eventElementPartPatterns);
    return eventTemplate;
  }

  public static Map<EventElementType, String> prepareMemoryTypeElementPartPatterns() {
    final Map<EventElementType, String> eventElementPartPatterns = new HashMap<>();

    eventElementPartPatterns.put(EventElementType.THREAD_IDENTIFIER, "([0-9a-f]){6,}+");
    eventElementPartPatterns.put(EventElementType.REGISTER, "[a-z0-9]{2,}+");
    eventElementPartPatterns.put(EventElementType.REGISTER_TYPE, "(obj|pri){1}");
    eventElementPartPatterns.put(EventElementType.REGISTER_VALUE, "[a-f0-9]{1,}+");

    return eventElementPartPatterns;
  }

  @Test
  public void memoryTypeLineRecognize00Pri() {
    this.messageAboutTest();

    final List<EventTemplate> eventMemory = new ArrayList<>();
    eventMemory.add(memoryTypeEventTemplate);
    final EventRecognizer eventRecognizer = new EventRecognizer(eventMemory);
    Logger.debug(eventRecognizer.toString());

    final String memoryLine = "[TID 002b0a] D/interpreter:      v1.pri = (i64) 0 | (f32) 0 | (f64) 0 | (hex) 0";
    Logger.debug("Line: \"%s\"", memoryLine);

    final ConcreteEvent concreteEvent = eventRecognizer.getEvent(memoryLine);
    Logger.debug("%s\n", concreteEvent);

    Logger.debug("REGISTER_TYPE: %s\n", concreteEvent.getElement(EventElementType.REGISTER_TYPE));

    assertEquals(EventType.MEMORY, concreteEvent.getEventType());
    assertEquals("002b0a", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
    assertEquals("v1", concreteEvent.getElement(EventElementType.REGISTER));
    assertEquals("pri", concreteEvent.getElement(EventElementType.REGISTER_TYPE));
    assertEquals("0", concreteEvent.getElement(EventElementType.REGISTER_VALUE));
  }

  @Test
  public void memoryTypeLineRecognize01Obj() {
    this.messageAboutTest();

    final List<EventTemplate> eventMemory = new ArrayList<>();
    eventMemory.add(memoryTypeEventTemplate);
    final EventRecognizer eventRecognizer = new EventRecognizer(eventMemory);
    Logger.debug(eventRecognizer.toString());

    final String memoryLine = "[TID 002b0a] D/interpreter:      v0.obj = 80080";
    Logger.debug("Line: \"%s\"", memoryLine);

    final ConcreteEvent concreteEvent = eventRecognizer.getEvent(memoryLine);
    Logger.debug("%s\n", concreteEvent);

    assertEquals(EventType.MEMORY, concreteEvent.getEventType());
    assertEquals("002b0a", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
    assertEquals("v0", concreteEvent.getElement(EventElementType.REGISTER));
    assertEquals("obj", concreteEvent.getElement(EventElementType.REGISTER_TYPE));
    assertEquals("80080", concreteEvent.getElement(EventElementType.REGISTER_VALUE));
  }

  // Instructions (examples 2)
  public static EventTemplate prepareInstructionPatternElements2() {
    final EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    final EventElementPattern eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    final EventElementPattern eventElementPc = new EventElementPattern(EventElementType.PC, "pc: 0x([a-f0-9]){2,}+ --->");
    final EventElementPattern eventElementInstruction = new EventElementPattern(EventElementType.INSTRUCTION, "\\S([a-z0-9|\\s|,]|[.]|[->]){2,}+");

    final List<EventElementPattern> eventElementTemplates = new ArrayList<>();
    eventElementTemplates.add(eventElementThreadId);
    eventElementTemplates.add(eventElementSource);
    eventElementTemplates.add(eventElementPc);
    eventElementTemplates.add(eventElementInstruction);

    final Map<EventElementType, String> eventElementPartPatterns = EventRecognizerTestCase.prepareInstructionElementPartPatterns2();

    final EventTemplate eventTemplate = new EventTemplate(EventType.INSTRUCTION, eventElementTemplates, eventElementPartPatterns);
    return eventTemplate;
  }

  public static Map<EventElementType, String> prepareInstructionElementPartPatterns2() {
    final Map<EventElementType, String> eventElementPartPatterns = new HashMap<>();

    eventElementPartPatterns.put(EventElementType.THREAD_IDENTIFIER, "([0-9a-f]){6,}+");
    eventElementPartPatterns.put(EventElementType.PC, "([a-f0-9]){2,}+");
    eventElementPartPatterns.put(EventElementType.INSTRUCTION, "([a-z0-9|\\s|,]|[.]|[->]){2,}+");

    return eventElementPartPatterns;
  }

  @Test
  public void instructionLineRecognize2() {
    this.messageAboutTest();

    final List<EventTemplate> eventInstruction = new ArrayList<>();
    eventInstruction.add(prepareInstructionPatternElements2());
    final EventRecognizer eventRecognizer = new EventRecognizer(eventInstruction);
    Logger.debug(eventRecognizer.toString());
    final String instructionLine = "[TID 09960f] D/interpreter:  pc: 0x7fae5b08c1fe ---> call.short id0, v1, v0";
    Logger.debug("Line: \"%s\"", instructionLine);

    final ConcreteEvent concreteEvent = eventRecognizer.getEvent(instructionLine);
    Logger.debug("%s\n", concreteEvent);

    assertEquals(EventType.INSTRUCTION, concreteEvent.getEventType());
    assertEquals("09960f", concreteEvent.getElement(EventElementType.THREAD_IDENTIFIER));
    assertEquals("7fae5b08c1fe", concreteEvent.getElement(EventElementType.PC));
    assertEquals("call.short id0, v1, v0", concreteEvent.getElement(EventElementType.INSTRUCTION));
  }

  // TODO: Remove
  private static String parseTemporaryLine(final String line, final String template) {
    Logger.debug("%s", line);
    Pattern p = Pattern.compile(template);
    Matcher m = p.matcher(line.subSequence(0, line.length()));
    m.find();
    Logger.debug("|%s| %s %s", m.group(), m.start(), m.end());
    //Logger.debug("|%s|", line.subSequence(36, line.length()));

    // while (m.find()) { Logger.debug("|%s| %s %s", m.group(), m.start(), m.end()); }
    // break;
    return m.group();
  }

  @Test
  public void simpleParse00() {
    final String result = parseTemporaryLine("[TID 09960f] D/interpreter:  pc: 0x7fae5b08c1fe ---> call.short id0, v1, v0", "pc: 0x([a-f0-9]){2,}+ --->");

    assertEquals("pc: 0x7fae5b08c1fe --->", result);
  }
}
