/*
 * Copyright 2022 ISP RAS (http://www.ispras.ru)
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License. You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License
 * is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
 * or implied. See the License for the specific language governing permissions and limitations under
 * the License.
 */

package ru.ispras.microtesk.tools.adapter.trace;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotEquals;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.junit.Before;
import org.junit.Test;

import ru.ispras.castle.util.Logger;

/**
 * @author <a href="mailto:protsenko@ispras.ru">Alexander Protsenko</a>
 */
public class EventTemplateTestCase {
  private EventTemplate eventElements;
  private EventTemplate eventMemory;
  private EventTemplate eventMemoryV2;
  private final List<EventTemplate> eventMemoryInstruction = new ArrayList<>();

  @Before
  public void prepareMockPatternElements() {
    final EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    final EventElementPattern eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");

    final List<EventElementPattern> eventElementTemplates = new ArrayList<>();
    eventElementTemplates.add(eventElementThreadId);
    eventElementTemplates.add(eventElementSource);

    this.eventElements = new EventTemplate(EventType.INSTRUCTION, eventElementTemplates, new HashMap<>());
  }

  @Before
  public void prepareMemoryPatternElements() {
    final EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    final EventElementPattern eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    final EventElementPattern eventElementRegister = new EventElementPattern(EventElementType.REGISTER, "\\s{2,}+[a-z0-9]{2,}+");
    final EventElementPattern eventElementRegisterValue = new EventElementPattern(EventElementType.REGISTER_VALUE, "[a-f0-9]{1,}+\\s\\|\\stag\\s=\\s[01]");
    // \[TID ([0-9a-f]){6,}+\](.){1,}?([a-zA-Z[//]]){2,}+:(.){1,}?\s{2,}+[a-z0-9]{2,}+(.){1,}?[a-f0-9]{1,}+\s\|\stag\s=\s[01]

    final List<EventElementPattern> eventElementTemplates = new ArrayList<>();
    eventElementTemplates.add(eventElementThreadId);
    eventElementTemplates.add(eventElementSource);
    eventElementTemplates.add(eventElementRegister);
    eventElementTemplates.add(eventElementRegisterValue);

    this.eventMemory = new EventTemplate(EventType.MEMORY, eventElementTemplates, new HashMap<>());
  }

  @Before
  public void prepareMemoryPatternV2Elements() {
    final EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    final EventElementPattern eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    final EventElementPattern eventElementRegister = new EventElementPattern(EventElementType.REGISTER, "\\s{2,}+[a-z0-9]{2,}+");
    final EventElementPattern eventElementRegisterValue = new EventElementPattern(EventElementType.REGISTER_VALUE, "[a-f0-9]{1,}+$");

    final List<EventElementPattern> eventElementTemplates = new ArrayList<>();
    eventElementTemplates.add(eventElementThreadId);
    eventElementTemplates.add(eventElementSource);
    eventElementTemplates.add(eventElementRegister);
    eventElementTemplates.add(eventElementRegisterValue);

    this.eventMemoryV2 = new EventTemplate(EventType.MEMORY, eventElementTemplates, new HashMap<>());
  }

  @Before
  public void prepareMemoryInstructionPatternElements() {
    // Memory
    EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    EventElementPattern eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    EventElementPattern eventElementRegister = new EventElementPattern(EventElementType.REGISTER, "\\s{2,}+[a-z0-9]{2,}+");
    EventElementPattern eventElementRegisterValue = new EventElementPattern(EventElementType.REGISTER_VALUE, "[a-f0-9]{1,}+$");

    List<EventElementPattern> eventElementTemplatesMemory = new ArrayList<>();
    eventElementTemplatesMemory.add(eventElementThreadId);
    eventElementTemplatesMemory.add(eventElementSource);
    eventElementTemplatesMemory.add(eventElementRegister);
    eventElementTemplatesMemory.add(eventElementRegisterValue);

    final EventTemplate eventTemplate = new EventTemplate(EventType.MEMORY, eventElementTemplatesMemory, new HashMap<>());
    this.eventMemoryInstruction.add(eventTemplate);

    // Instruction
    eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    eventElementSource = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    final EventElementPattern eventElementPc = new EventElementPattern(EventElementType.PC, "([a-f0-9]){2,}+:");
    final EventElementPattern eventElementInstruction = new EventElementPattern(EventElementType.INSTRUCTION, "\\S([a-z0-9|\\s|,]|[.]|[->]){2,}+");

    final List<EventElementPattern> eventElementTemplatesInstruction = new ArrayList<>();
    eventElementTemplatesInstruction.add(eventElementThreadId);
    eventElementTemplatesInstruction.add(eventElementSource);
    eventElementTemplatesInstruction.add(eventElementPc);
    eventElementTemplatesInstruction.add(eventElementInstruction);
    EventTemplate eventTemplateInstruction = new EventTemplate(EventType.INSTRUCTION, eventElementTemplatesInstruction, new HashMap<>());
    this.eventMemoryInstruction.add(eventTemplateInstruction);
  }

  @Test
  public void simpleEmpty() {
    final EventTemplate eventTemplate = new EventTemplate(EventType.INFO, new ArrayList<>(), new HashMap<>());

    final String lineToCheck = "[TID 00273b] D/interpreter: ";
    Logger.message("Test for: %s vs. %s", eventTemplate, lineToCheck);

    final boolean check = eventTemplate.checkLineEventType(lineToCheck);

    assertEquals(true, check);
  }

  @Test
  public void simpleInstructionTrue() {
    final String lineToCheck = "[TID 00273b] D/interpreter: ";
    Logger.message("Test 'true' for: %s vs. %s", eventElements, lineToCheck);

    final boolean check = eventElements.checkLineEventType(lineToCheck);

    assertEquals(true, check);
    assertEquals(EventType.INSTRUCTION, eventElements.getEventType());
  }

  @Test
  public void simpleInstructionFalse() {
    final String lineToCheck = "[TID 00273b]";
    Logger.message("Test 'false' for: %s vs. %s", eventElements, lineToCheck);

    final boolean check = eventElements.checkLineEventType(lineToCheck);

    assertNotEquals(true, check);
  }

  @Test
  public void simpleInstructionFalse2() {
    final String lineToCheck = "[TID 00273b] D/interpreter[: ";
    Logger.message("Test 'false' for: %s vs. %s", eventElements, lineToCheck);

    final boolean check = eventElements.checkLineEventType(lineToCheck);

    assertNotEquals(true, check);
  }

  @Test
  public void simpleMemoryTrue() {
    final String lineToCheck = "[TID 002984] D/interpreter:      v0.pri = (i64) 0 | (f32) 0 | (f64) 0 | (hex) 0 | tag = 0";
    Logger.message("Test 'true' for: %s vs. %s", eventMemory, lineToCheck);

    final boolean check = eventMemory.checkLineEventType(lineToCheck);

    assertEquals(true, check);
  }

  @Test
  public void simpleMemoryV2True() {
    final String lineToCheck = "[TID 006b07] D/interpreter:      v0.pri = (i64) 0 | (f32) 0 | (f64) 0 | (hex) 0";
    Logger.message("Test 'true' for: %s vs. %s", eventMemoryV2, lineToCheck);

    final boolean check = eventMemoryV2.checkLineEventType(lineToCheck);

    assertEquals(true, check);
    assertEquals(EventType.MEMORY, eventMemoryV2.getEventType());
  }

  @Test
  public void simpleMemoryV2True2() {
    final String lineToCheck = "[TID 006b07] D/interpreter:      acc.pri = (i64) 255 | (f32) 3.57331e-43 | (f64) 1.25987e-321 | (hex) ff";
    Logger.message("Test 'true' for: %s vs. %s", eventMemoryV2, lineToCheck);

    final boolean check = eventMemoryV2.checkLineEventType(lineToCheck);

    assertEquals(true, check);
    assertEquals(EventType.MEMORY, eventMemoryV2.getEventType());
  }

  private EventTemplate eventTemplateFrameCreation;

  @Before
  public void prepareMockEventTemplateFrameCreation() {
    final EventElementPattern eventElementThreadId = new EventElementPattern(EventElementType.THREAD_IDENTIFIER, "\\[TID ([0-9a-f]){6,}+\\]");
    final EventElementPattern eventElementSource   = new EventElementPattern(EventElementType.SOURCE, "([a-zA-Z[//]]){2,}+:");
    final EventElementPattern eventElementFrame    = new EventElementPattern(EventElementType.FRAME_ENTRY, "(.*?)(\\bEntry\\b)(.*)$");

    final List<EventElementPattern> eventElementTemplates = new ArrayList<>();
    eventElementTemplates.add(eventElementThreadId);
    eventElementTemplates.add(eventElementSource);
    eventElementTemplates.add(eventElementFrame);

    this.eventTemplateFrameCreation = new EventTemplate(EventType.FRAME_ENTRY, eventElementTemplates, new HashMap<>());
  }

  @Test
  public void frameCreationEntry() {
    final String lineToCheck = "[TID 002b0a] D/interpreter: Entry: class8060::method140";
    Logger.message("Test 'true' for: %s vs. %s", eventTemplateFrameCreation, lineToCheck);

    final boolean check = eventTemplateFrameCreation.checkLineEventType(lineToCheck);

    assertEquals(true, check);
  }
}
